# stdlib
import json
from typing import Any
from typing import Dict
from typing import List
from typing import Optional

# third party
from google.protobuf.reflection import GeneratedProtocolMessageType
from typing_extensions import final

# relative
from ...... import serialize
from ......proto.grid.messages.request_messages_pb2 import (
    CreateBudgetRequestMessage as CreateBudgetRequestMessage_PB,
)
from ......proto.grid.messages.request_messages_pb2 import (
    CreateRequestMessage as CreateRequestMessage_PB,
)
from ......proto.grid.messages.request_messages_pb2 import (
    CreateRequestResponse as CreateRequestResponse_PB,
)
from ......proto.grid.messages.request_messages_pb2 import (
    DeleteRequestMessage as DeleteRequestMessage_PB,
)
from ......proto.grid.messages.request_messages_pb2 import (
    DeleteRequestResponse as DeleteRequestResponse_PB,
)
from ......proto.grid.messages.request_messages_pb2 import (
    GetAllRequestsMessage as GetAllRequestsMessage_PB,
)
from ......proto.grid.messages.request_messages_pb2 import (
    GetAllRequestsResponseMessage as GetAllRequestsResponseMessage_PB,
)
from ......proto.grid.messages.request_messages_pb2 import (
    GetBudgetRequestsMessage as GetBudgetRequestsMessage_PB,
)
from ......proto.grid.messages.request_messages_pb2 import (
    GetBudgetRequestsResponse as GetBudgetRequestsResponse_PB,
)
from ......proto.grid.messages.request_messages_pb2 import (
    GetRequestMessage as GetRequestMessage_PB,
)
from ......proto.grid.messages.request_messages_pb2 import (
    GetRequestResponse as GetRequestResponse_PB,
)
from ......proto.grid.messages.request_messages_pb2 import (
    GetRequestsMessage as GetRequestsMessage_PB,
)
from ......proto.grid.messages.request_messages_pb2 import (
    GetRequestsResponse as GetRequestsResponse_PB,
)
from ......proto.grid.messages.request_messages_pb2 import (
    UpdateRequestMessage as UpdateRequestMessage_PB,
)
from ......proto.grid.messages.request_messages_pb2 import (
    UpdateRequestResponse as UpdateRequestResponse_PB,
)
from .....common.message import ImmediateSyftMessageWithReply
from .....common.message import ImmediateSyftMessageWithoutReply
from .....common.serde.deserialize import _deserialize
from .....common.serde.serializable import serializable
from .....common.uid import UID
from .....io.address import Address
from ..request_receiver.request_receiver_messages import RequestMessage


@serializable()
@final
class CreateRequestMessage(ImmediateSyftMessageWithReply):
    def __init__(
        self,
        address: Address,
        content: Dict,
        reply_to: Address,
        msg_id: Optional[UID] = None,
    ):
        super().__init__(address=address, msg_id=msg_id, reply_to=reply_to)
        self.content = content

    def _object2proto(self) -> CreateRequestMessage_PB:
        """Returns a protobuf serialization of self.
        As a requirement of all objects which inherit from Serializable,
        this method transforms the current object into the corresponding
        Protobuf object so that it can be further serialized.
        :return: returns a protobuf object
        :rtype: CreateRequestMessage_PB
        .. note::
            This method is purely an internal method. Please use serialize(object) or one of
            the other public serialization methods if you wish to serialize an
            object.
        """
        return CreateRequestMessage_PB(
            msg_id=serialize(self.id),
            address=serialize(self.address),
            content=json.dumps(self.content),
            reply_to=serialize(self.reply_to),
        )

    @staticmethod
    def _proto2object(
        proto: CreateRequestMessage_PB,
    ) -> "CreateRequestMessage":
        """Creates a CreateRequestMessage from a protobuf
        As a requirement of all objects which inherit from Serializable,
        this method transforms a protobuf object into an instance of this class.
        :return: returns an instance of SignalingOfferMessage
        :rtype: CreateRequestMessage
        .. note::
            This method is purely an internal method. Please use syft.deserialize()
            if you wish to deserialize an object.
        """

        return CreateRequestMessage(
            msg_id=_deserialize(blob=proto.msg_id),
            address=_deserialize(blob=proto.address),
            content=json.loads(proto.content),
            reply_to=_deserialize(blob=proto.reply_to),
        )

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        """Return the type of protobuf object which stores a class of this type
        As a part of serialization and deserialization, we need the ability to
        lookup the protobuf object type directly from the object type. This
        static method allows us to do this.
        Importantly, this method is also used to create the reverse lookup ability within
        the metaclass of Serializable. In the metaclass, it calls this method and then
        it takes whatever type is returned from this method and adds an attribute to it
        with the type of this class attached to it. See the MetaSerializable class for
        details.
        :return: the type of protobuf object which corresponds to this class.
        :rtype: GeneratedProtocolMessageType
        """

        return CreateRequestMessage_PB


@serializable()
@final
class CreateBudgetRequestMessage(ImmediateSyftMessageWithoutReply):
    def __init__(
        self,
        address: Address,
        budget: float,
        reason: str,
        msg_id: Optional[UID] = None,
    ):
        super().__init__(address=address, msg_id=msg_id)
        self.budget = budget
        self.reason = reason

    def _object2proto(self) -> CreateBudgetRequestMessage_PB:
        """Returns a protobuf serialization of self.
        As a requirement of all objects which inherit from Serializable,
        this method transforms the current object into the corresponding
        Protobuf object so that it can be further serialized.
        :return: returns a protobuf object
        :rtype: CreateBudgetRequestMessage_PB
        .. note::
            This method is purely an internal method. Please use serialize(object) or one of
            the other public serialization methods if you wish to serialize an
            object.
        """
        return CreateBudgetRequestMessage_PB(
            msg_id=serialize(self.id),
            address=serialize(self.address),
            budget=self.budget,
            reason=self.reason,
        )

    @staticmethod
    def _proto2object(
        proto: CreateBudgetRequestMessage_PB,
    ) -> "CreateBudgetRequestMessage":
        """Creates a CreateRequestMessage from a protobuf
        As a requirement of all objects which inherit from Serializable,
        this method transforms a protobuf object into an instance of this class.
        :return: returns an instance of SignalingOfferMessage
        :rtype: CreateRequestMessage
        .. note::
            This method is purely an internal method. Please use syft.deserialize()
            if you wish to deserialize an object.
        """

        return CreateBudgetRequestMessage(
            msg_id=_deserialize(blob=proto.msg_id),
            address=_deserialize(blob=proto.address),
            budget=proto.budget,
            reason=proto.reason,
        )

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        """Return the type of protobuf object which stores a class of this type
        As a part of serialization and deserialization, we need the ability to
        lookup the protobuf object type directly from the object type. This
        static method allows us to do this.
        Importantly, this method is also used to create the reverse lookup ability within
        the metaclass of Serializable. In the metaclass, it calls this method and then
        it takes whatever type is returned from this method and adds an attribute to it
        with the type of this class attached to it. See the MetaSerializable class for
        details.
        :return: the type of protobuf object which corresponds to this class.
        :rtype: GeneratedProtocolMessageType
        """

        return CreateBudgetRequestMessage_PB


@serializable()
@final
class CreateRequestResponse(ImmediateSyftMessageWithoutReply):
    def __init__(
        self,
        address: Address,
        status_code: int,
        content: Dict,
        msg_id: Optional[UID] = None,
    ):
        super().__init__(address=address, msg_id=msg_id)
        self.status_code = status_code
        self.content = content

    def _object2proto(self) -> CreateRequestResponse_PB:
        """Returns a protobuf serialization of self.
        As a requirement of all objects which inherit from Serializable,
        this method transforms the current object into the corresponding
        Protobuf object so that it can be further serialized.
        :return: returns a protobuf object
        :rtype: SignalingOfferMessage_PB
        .. note::
            This method is purely an internal method. Please use serialize(object) or one of
            the other public serialization methods if you wish to serialize an
            object.
        """
        return CreateRequestResponse_PB(
            msg_id=serialize(self.id),
            address=serialize(self.address),
            status_code=self.status_code,
            content=json.dumps(self.content),
        )

    @staticmethod
    def _proto2object(
        proto: CreateRequestResponse_PB,
    ) -> "CreateRequestResponse":
        """Creates a SignalingOfferMessage from a protobuf
        As a requirement of all objects which inherit from Serializable,
        this method transforms a protobuf object into an instance of this class.
        :return: returns an instance of SignalingOfferMessage
        :rtype: SignalingOfferMessage
        .. note::
            This method is purely an internal method. Please use syft.deserialize()
            if you wish to deserialize an object.
        """

        return CreateRequestResponse(
            msg_id=_deserialize(blob=proto.msg_id),
            address=_deserialize(blob=proto.address),
            status_code=proto.status_code,
            content=json.loads(proto.content),
        )

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        """Return the type of protobuf object which stores a class of this type
        As a part of serialization and deserialization, we need the ability to
        lookup the protobuf object type directly from the object type. This
        static method allows us to do this.
        Importantly, this method is also used to create the reverse lookup ability within
        the metaclass of Serializable. In the metaclass, it calls this method and then
        it takes whatever type is returned from this method and adds an attribute to it
        with the type of this class attached to it. See the MetaSerializable class for
        details.
        :return: the type of protobuf object which corresponds to this class.
        :rtype: GeneratedProtocolMessageType
        """

        return CreateRequestResponse_PB


@serializable()
@final
class GetRequestMessage(ImmediateSyftMessageWithReply):
    def __init__(
        self,
        address: Address,
        request_id: str,
        reply_to: Address,
        msg_id: Optional[UID] = None,
    ):
        super().__init__(address=address, msg_id=msg_id, reply_to=reply_to)
        self.request_id = request_id

    def _object2proto(self) -> GetRequestMessage_PB:
        """Returns a protobuf serialization of self.
        As a requirement of all objects which inherit from Serializable,
        this method transforms the current object into the corresponding
        Protobuf object so that it can be further serialized.
        :return: returns a protobuf object
        :rtype: GetRequestMessage_PB
        .. note::
            This method is purely an internal method. Please use serialize(object) or one of
            the other public serialization methods if you wish to serialize an
            object.
        """
        return GetRequestMessage_PB(
            msg_id=serialize(self.id),
            address=serialize(self.address),
            request_id=json.dumps(self.request_id),
            reply_to=serialize(self.reply_to),
        )

    @staticmethod
    def _proto2object(
        proto: GetRequestMessage_PB,
    ) -> "GetRequestMessage":
        """Creates a GetRequestMessage from a protobuf
        As a requirement of all objects which inherit from Serializable,
        this method transforms a protobuf object into an instance of this class.
        :return: returns an instance of SignalingOfferMessage
        :rtype: GetRequestMessage
        .. note::
            This method is purely an internal method. Please use syft.deserialize()
            if you wish to deserialize an object.
        """

        return GetRequestMessage(
            msg_id=_deserialize(blob=proto.msg_id),
            address=_deserialize(blob=proto.address),
            request_id=json.loads(proto.request_id),
            reply_to=_deserialize(blob=proto.reply_to),
        )

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        """Return the type of protobuf object which stores a class of this type
        As a part of serialization and deserialization, we need the ability to
        lookup the protobuf object type directly from the object type. This
        static method allows us to do this.
        Importantly, this method is also used to create the reverse lookup ability within
        the metaclass of Serializable. In the metaclass, it calls this method and then
        it takes whatever type is returned from this method and adds an attribute to it
        with the type of this class attached to it. See the MetaSerializable class for
        details.
        :return: the type of protobuf object which corresponds to this class.
        :rtype: GeneratedProtocolMessageType
        """

        return GetRequestMessage_PB


@serializable()
@final
class GetRequestResponse(ImmediateSyftMessageWithoutReply):
    def __init__(
        self,
        address: Address,
        status_code: int,
        request_id: Dict[str, Any],
        msg_id: Optional[UID] = None,
    ):
        super().__init__(address=address, msg_id=msg_id)
        self.status_code = status_code
        self.request_id = request_id

    def _object2proto(self) -> GetRequestResponse_PB:
        """Returns a protobuf serialization of self.
        As a requirement of all objects which inherit from Serializable,
        this method transforms the current object into the corresponding
        Protobuf object so that it can be further serialized.
        :return: returns a protobuf object
        :rtype: SignalingOfferMessage_PB
        .. note::
            This method is purely an internal method. Please use serialize(object) or one of
            the other public serialization methods if you wish to serialize an
            object.
        """
        return GetRequestResponse_PB(
            msg_id=serialize(self.id),
            address=serialize(self.address),
            status_code=self.status_code,
            request_id=json.dumps(self.request_id),
        )

    @staticmethod
    def _proto2object(
        proto: GetRequestResponse_PB,
    ) -> "GetRequestResponse":
        """Creates a SignalingOfferMessage from a protobuf
        As a requirement of all objects which inherit from Serializable,
        this method transforms a protobuf object into an instance of this class.
        :return: returns an instance of SignalingOfferMessage
        :rtype: SignalingOfferMessage
        .. note::
            This method is purely an internal method. Please use syft.deserialize()
            if you wish to deserialize an object.
        """

        return GetRequestResponse(
            msg_id=_deserialize(blob=proto.msg_id),
            address=_deserialize(blob=proto.address),
            status_code=proto.status_code,
            request_id=json.loads(proto.request_id),
        )

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        """Return the type of protobuf object which stores a class of this type
        As a part of serialization and deserialization, we need the ability to
        lookup the protobuf object type directly from the object type. This
        static method allows us to do this.
        Importantly, this method is also used to create the reverse lookup ability within
        the metaclass of Serializable. In the metaclass, it calls this method and then
        it takes whatever type is returned from this method and adds an attribute to it
        with the type of this class attached to it. See the MetaSerializable class for
        details.
        :return: the type of protobuf object which corresponds to this class.
        :rtype: GeneratedProtocolMessageType
        """

        return GetRequestResponse_PB


@serializable()
@final
class GetRequestsMessage(ImmediateSyftMessageWithReply):
    def __init__(
        self,
        address: Address,
        reply_to: Address,
        msg_id: Optional[UID] = None,
    ):
        super().__init__(address=address, msg_id=msg_id, reply_to=reply_to)

    def _object2proto(self) -> GetRequestsMessage_PB:
        """Returns a protobuf serialization of self.
        As a requirement of all objects which inherit from Serializable,
        this method transforms the current object into the corresponding
        Protobuf object so that it can be further serialized.
        :return: returns a protobuf object
        :rtype: GetRequestsMessage_PB
        .. note::
            This method is purely an internal method. Please use serialize(object) or one of
            the other public serialization methods if you wish to serialize an
            object.
        """
        return GetRequestsMessage_PB(
            msg_id=serialize(self.id),
            address=serialize(self.address),
            reply_to=serialize(self.reply_to),
        )

    @staticmethod
    def _proto2object(
        proto: GetRequestsMessage_PB,
    ) -> "GetRequestsMessage":
        """Creates a GetRequestsMessage from a protobuf
        As a requirement of all objects which inherit from Serializable,
        this method transforms a protobuf object into an instance of this class.
        :return: returns an instance of SignalingOfferMessage
        :rtype: GetRequestsMessage
        .. note::
            This method is purely an internal method. Please use syft.deserialize()
            if you wish to deserialize an object.
        """

        return GetRequestsMessage(
            msg_id=_deserialize(blob=proto.msg_id),
            address=_deserialize(blob=proto.address),
            reply_to=_deserialize(blob=proto.reply_to),
        )

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        """Return the type of protobuf object which stores a class of this type
        As a part of serialization and deserialization, we need the ability to
        lookup the protobuf object type directly from the object type. This
        static method allows us to do this.
        Importantly, this method is also used to create the reverse lookup ability within
        the metaclass of Serializable. In the metaclass, it calls this method and then
        it takes whatever type is returned from this method and adds an attribute to it
        with the type of this class attached to it. See the MetaSerializable class for
        details.
        :return: the type of protobuf object which corresponds to this class.
        :rtype: GeneratedProtocolMessageType
        """

        return GetRequestsMessage_PB


@serializable()
@final
class GetRequestsResponse(ImmediateSyftMessageWithoutReply):
    def __init__(
        self,
        address: Address,
        status_code: int,
        content: List[Dict],
        msg_id: Optional[UID] = None,
    ):
        super().__init__(address=address, msg_id=msg_id)
        self.status_code = status_code
        self.content = content

    def _object2proto(self) -> GetRequestsResponse_PB:
        """Returns a protobuf serialization of self.
        As a requirement of all objects which inherit from Serializable,
        this method transforms the current object into the corresponding
        Protobuf object so that it can be further serialized.
        :return: returns a protobuf object
        :rtype: SignalingOfferMessage_PB
        .. note::
            This method is purely an internal method. Please use serialize(object) or one of
            the other public serialization methods if you wish to serialize an
            object.
        """
        msg = GetRequestsResponse_PB(
            msg_id=serialize(self.id),
            address=serialize(self.address),
            status_code=self.status_code,
        )

        for content in self.content:
            msg.content.append(serialize(content))

        return msg

    @staticmethod
    def _proto2object(
        proto: GetRequestsResponse_PB,
    ) -> "GetRequestsResponse":
        """Creates a SignalingOfferMessage from a protobuf
        As a requirement of all objects which inherit from Serializable,
        this method transforms a protobuf object into an instance of this class.
        :return: returns an instance of SignalingOfferMessage
        :rtype: SignalingOfferMessage
        .. note::
            This method is purely an internal method. Please use syft.deserialize()
            if you wish to deserialize an object.
        """
        return GetRequestsResponse(
            msg_id=_deserialize(blob=proto.msg_id),
            address=_deserialize(blob=proto.address),
            status_code=proto.status_code,
            content=[_deserialize(content) for content in proto.content],
        )

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        """Return the type of protobuf object which stores a class of this type
        As a part of serialization and deserialization, we need the ability to
        lookup the protobuf object type directly from the object type. This
        static method allows us to do this.
        Importantly, this method is also used to create the reverse lookup ability within
        the metaclass of Serializable. In the metaclass, it calls this method and then
        it takes whatever type is returned from this method and adds an attribute to it
        with the type of this class attached to it. See the MetaSerializable class for
        details.
        :return: the type of protobuf object which corresponds to this class.
        :rtype: GeneratedProtocolMessageType
        """

        return GetRequestsResponse_PB


@serializable()
@final
class GetBudgetRequestsMessage(ImmediateSyftMessageWithReply):
    def __init__(
        self,
        address: Address,
        reply_to: Address,
        msg_id: Optional[UID] = None,
    ):
        super().__init__(address=address, msg_id=msg_id, reply_to=reply_to)

    def _object2proto(self) -> GetBudgetRequestsMessage_PB:
        """Returns a protobuf serialization of self.
        As a requirement of all objects which inherit from Serializable,
        this method transforms the current object into the corresponding
        Protobuf object so that it can be further serialized.
        :return: returns a protobuf object
        :rtype: GetRequestsMessage_PB
        .. note::
            This method is purely an internal method. Please use serialize(object) or one of
            the other public serialization methods if you wish to serialize an
            object.
        """
        return GetBudgetRequestsMessage_PB(
            msg_id=serialize(self.id),
            address=serialize(self.address),
            reply_to=serialize(self.reply_to),
        )

    @staticmethod
    def _proto2object(
        proto: GetBudgetRequestsMessage_PB,
    ) -> "GetBudgetRequestsMessage":
        """Creates a GetRequestsMessage from a protobuf
        As a requirement of all objects which inherit from Serializable,
        this method transforms a protobuf object into an instance of this class.
        :return: returns an instance of SignalingOfferMessage
        :rtype: GetRequestsMessage
        .. note::
            This method is purely an internal method. Please use syft.deserialize()
            if you wish to deserialize an object.
        """

        return GetBudgetRequestsMessage(
            msg_id=_deserialize(blob=proto.msg_id),
            address=_deserialize(blob=proto.address),
            reply_to=_deserialize(blob=proto.reply_to),
        )

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        """Return the type of protobuf object which stores a class of this type
        As a part of serialization and deserialization, we need the ability to
        lookup the protobuf object type directly from the object type. This
        static method allows us to do this.
        Importantly, this method is also used to create the reverse lookup ability within
        the metaclass of Serializable. In the metaclass, it calls this method and then
        it takes whatever type is returned from this method and adds an attribute to it
        with the type of this class attached to it. See the MetaSerializable class for
        details.
        :return: the type of protobuf object which corresponds to this class.
        :rtype: GeneratedProtocolMessageType
        """

        return GetBudgetRequestsMessage_PB


@serializable()
@final
class GetBudgetRequestsResponse(ImmediateSyftMessageWithoutReply):
    def __init__(
        self,
        address: Address,
        content: List[Dict],
        msg_id: Optional[UID] = None,
    ):
        super().__init__(address=address, msg_id=msg_id)
        self.content = content

    def _object2proto(self) -> GetBudgetRequestsResponse_PB:
        """Returns a protobuf serialization of self.
        As a requirement of all objects which inherit from Serializable,
        this method transforms the current object into the corresponding
        Protobuf object so that it can be further serialized.
        :return: returns a protobuf object
        :rtype: SignalingOfferMessage_PB
        .. note::
            This method is purely an internal method. Please use serialize(object) or one of
            the other public serialization methods if you wish to serialize an
            object.
        """
        msg = GetBudgetRequestsResponse_PB(
            msg_id=serialize(self.id),
            address=serialize(self.address),
        )

        for content in self.content:
            msg.content.append(serialize(content))

        return msg

    @staticmethod
    def _proto2object(
        proto: GetBudgetRequestsResponse_PB,
    ) -> "GetBudgetRequestsResponse":
        """Creates a SignalingOfferMessage from a protobuf
        As a requirement of all objects which inherit from Serializable,
        this method transforms a protobuf object into an instance of this class.
        :return: returns an instance of SignalingOfferMessage
        :rtype: SignalingOfferMessage
        .. note::
            This method is purely an internal method. Please use syft.deserialize()
            if you wish to deserialize an object.
        """
        return GetBudgetRequestsResponse(
            msg_id=_deserialize(blob=proto.msg_id),
            address=_deserialize(blob=proto.address),
            content=[_deserialize(content) for content in proto.content],
        )

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        """Return the type of protobuf object which stores a class of this type
        As a part of serialization and deserialization, we need the ability to
        lookup the protobuf object type directly from the object type. This
        static method allows us to do this.
        Importantly, this method is also used to create the reverse lookup ability within
        the metaclass of Serializable. In the metaclass, it calls this method and then
        it takes whatever type is returned from this method and adds an attribute to it
        with the type of this class attached to it. See the MetaSerializable class for
        details.
        :return: the type of protobuf object which corresponds to this class.
        :rtype: GeneratedProtocolMessageType
        """

        return GetBudgetRequestsResponse_PB


@serializable()
@final
class UpdateRequestMessage(ImmediateSyftMessageWithReply):
    def __init__(
        self,
        address: Address,
        request_id: str,
        status: str,
        reply_to: Address,
        msg_id: Optional[UID] = None,
    ):
        super().__init__(address=address, msg_id=msg_id, reply_to=reply_to)
        self.request_id = request_id
        self.status = status

    def _object2proto(self) -> UpdateRequestMessage_PB:
        """Returns a protobuf serialization of self.
        As a requirement of all objects which inherit from Serializable,
        this method transforms the current object into the corresponding
        Protobuf object so that it can be further serialized.
        :return: returns a protobuf object
        :rtype: UpdateRequestMessage_PB
        .. note::
            This method is purely an internal method. Please use serialize(object) or one of
            the other public serialization methods if you wish to serialize an
            object.
        """
        return UpdateRequestMessage_PB(
            msg_id=serialize(self.id),
            address=serialize(self.address),
            request_id=json.dumps(self.request_id),
            status=json.dumps(self.status),
            reply_to=serialize(self.reply_to),
        )

    @staticmethod
    def _proto2object(
        proto: UpdateRequestMessage_PB,
    ) -> "UpdateRequestMessage":
        """Creates a UpdateRequestMessage from a protobuf
        As a requirement of all objects which inherit from Serializable,
        this method transforms a protobuf object into an instance of this class.
        :return: returns an instance of SignalingOfferMessage
        :rtype: UpdateRequestMessage
        .. note::
            This method is purely an internal method. Please use syft.deserialize()
            if you wish to deserialize an object.
        """

        return UpdateRequestMessage(
            msg_id=_deserialize(blob=proto.msg_id),
            address=_deserialize(blob=proto.address),
            request_id=json.loads(proto.request_id),
            status=json.loads(proto.status),
            reply_to=_deserialize(blob=proto.reply_to),
        )

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        """Return the type of protobuf object which stores a class of this type
        As a part of serialization and deserialization, we need the ability to
        lookup the protobuf object type directly from the object type. This
        static method allows us to do this.
        Importantly, this method is also used to create the reverse lookup ability within
        the metaclass of Serializable. In the metaclass, it calls this method and then
        it takes whatever type is returned from this method and adds an attribute to it
        with the type of this class attached to it. See the MetaSerializable class for
        details.
        :return: the type of protobuf object which corresponds to this class.
        :rtype: GeneratedProtocolMessageType
        """

        return UpdateRequestMessage_PB


@serializable()
@final
class UpdateRequestResponse(ImmediateSyftMessageWithoutReply):
    def __init__(
        self,
        address: Address,
        status_code: int,
        status: str,
        request_id: str,
        msg_id: Optional[UID] = None,
    ):
        super().__init__(address=address, msg_id=msg_id)
        self.status_code = status_code
        self.status = status
        self.request_id = request_id

    def _object2proto(self) -> UpdateRequestResponse_PB:
        """Returns a protobuf serialization of self.
        As a requirement of all objects which inherit from Serializable,
        this method transforms the current object into the corresponding
        Protobuf object so that it can be further serialized.
        :return: returns a protobuf object
        :rtype: SignalingOfferMessage_PB
        .. note::
            This method is purely an internal method. Please use serialize(object) or one of
            the other public serialization methods if you wish to serialize an
            object.
        """
        return UpdateRequestResponse_PB(
            msg_id=serialize(self.id),
            address=serialize(self.address),
            status_code=self.status_code,
            status=json.dumps(self.status),
            request_id=json.dumps(self.request_id),
        )

    @staticmethod
    def _proto2object(
        proto: UpdateRequestResponse_PB,
    ) -> "UpdateRequestResponse":
        """Creates a SignalingOfferMessage from a protobuf
        As a requirement of all objects which inherit from Serializable,
        this method transforms a protobuf object into an instance of this class.
        :return: returns an instance of SignalingOfferMessage
        :rtype: SignalingOfferMessage
        .. note::
            This method is purely an internal method. Please use syft.deserialize()
            if you wish to deserialize an object.
        """

        return UpdateRequestResponse(
            msg_id=_deserialize(blob=proto.msg_id),
            address=_deserialize(blob=proto.address),
            status_code=proto.status_code,
            request_id=json.loads(proto.request_id),
            status=json.loads(proto.status),
        )

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        """Return the type of protobuf object which stores a class of this type
        As a part of serialization and deserialization, we need the ability to
        lookup the protobuf object type directly from the object type. This
        static method allows us to do this.
        Importantly, this method is also used to create the reverse lookup ability within
        the metaclass of Serializable. In the metaclass, it calls this method and then
        it takes whatever type is returned from this method and adds an attribute to it
        with the type of this class attached to it. See the MetaSerializable class for
        details.
        :return: the type of protobuf object which corresponds to this class.
        :rtype: GeneratedProtocolMessageType
        """

        return UpdateRequestResponse_PB


@serializable()
@final
class DeleteRequestMessage(ImmediateSyftMessageWithReply):
    def __init__(
        self,
        address: Address,
        request_id: str,
        reply_to: Address,
        msg_id: Optional[UID] = None,
    ):
        super().__init__(address=address, msg_id=msg_id, reply_to=reply_to)
        self.request_id = request_id

    def _object2proto(self) -> DeleteRequestMessage_PB:
        """Returns a protobuf serialization of self.
        As a requirement of all objects which inherit from Serializable,
        this method transforms the current object into the corresponding
        Protobuf object so that it can be further serialized.
        :return: returns a protobuf object
        :rtype: DeleteRequestMessage_PB
        .. note::
            This method is purely an internal method. Please use serialize(object) or one of
            the other public serialization methods if you wish to serialize an
            object.
        """
        return DeleteRequestMessage_PB(
            msg_id=serialize(self.id),
            address=serialize(self.address),
            request_id=json.dumps(self.request_id),
            reply_to=serialize(self.reply_to),
        )

    @staticmethod
    def _proto2object(
        proto: DeleteRequestMessage_PB,
    ) -> "DeleteRequestMessage":
        """Creates a DeleteRequestMessage from a protobuf
        As a requirement of all objects which inherit from Serializable,
        this method transforms a protobuf object into an instance of this class.
        :return: returns an instance of SignalingOfferMessage
        :rtype: DeleteRequestMessage
        .. note::
            This method is purely an internal method. Please use syft.deserialize()
            if you wish to deserialize an object.
        """

        return DeleteRequestMessage(
            msg_id=_deserialize(blob=proto.msg_id),
            address=_deserialize(blob=proto.address),
            request_id=json.loads(proto.request_id),
            reply_to=_deserialize(blob=proto.reply_to),
        )

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        """Return the type of protobuf object which stores a class of this type
        As a part of serialization and deserialization, we need the ability to
        lookup the protobuf object type directly from the object type. This
        static method allows us to do this.
        Importantly, this method is also used to create the reverse lookup ability within
        the metaclass of Serializable. In the metaclass, it calls this method and then
        it takes whatever type is returned from this method and adds an attribute to it
        with the type of this class attached to it. See the MetaSerializable class for
        details.
        :return: the type of protobuf object which corresponds to this class.
        :rtype: GeneratedProtocolMessageType
        """

        return DeleteRequestMessage_PB


@serializable()
@final
class DeleteRequestResponse(ImmediateSyftMessageWithoutReply):
    def __init__(
        self,
        address: Address,
        status_code: int,
        request_id: str,
        msg_id: Optional[UID] = None,
    ):
        super().__init__(address=address, msg_id=msg_id)
        self.status_code = status_code
        self.request_id = request_id

    def _object2proto(self) -> DeleteRequestResponse_PB:
        """Returns a protobuf serialization of self.
        As a requirement of all objects which inherit from Serializable,
        this method transforms the current object into the corresponding
        Protobuf object so that it can be further serialized.
        :return: returns a protobuf object
        :rtype: SignalingOfferMessage_PB
        .. note::
            This method is purely an internal method. Please use serialize(object) or one of
            the other public serialization methods if you wish to serialize an
            object.
        """
        return DeleteRequestResponse_PB(
            msg_id=serialize(self.id),
            address=serialize(self.address),
            status_code=self.status_code,
            request_id=json.dumps(self.request_id),
        )

    @staticmethod
    def _proto2object(
        proto: DeleteRequestResponse_PB,
    ) -> "DeleteRequestResponse":
        """Creates a SignalingOfferMessage from a protobuf
        As a requirement of all objects which inherit from Serializable,
        this method transforms a protobuf object into an instance of this class.
        :return: returns an instance of SignalingOfferMessage
        :rtype: SignalingOfferMessage
        .. note::
            This method is purely an internal method. Please use syft.deserialize()
            if you wish to deserialize an object.
        """

        return DeleteRequestResponse(
            msg_id=_deserialize(blob=proto.msg_id),
            address=_deserialize(blob=proto.address),
            status_code=proto.status_code,
            request_id=json.loads(proto.request_id),
        )

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        """Return the type of protobuf object which stores a class of this type
        As a part of serialization and deserialization, we need the ability to
        lookup the protobuf object type directly from the object type. This
        static method allows us to do this.
        Importantly, this method is also used to create the reverse lookup ability within
        the metaclass of Serializable. In the metaclass, it calls this method and then
        it takes whatever type is returned from this method and adds an attribute to it
        with the type of this class attached to it. See the MetaSerializable class for
        details.
        :return: the type of protobuf object which corresponds to this class.
        :rtype: GeneratedProtocolMessageType
        """

        return DeleteRequestResponse_PB


@serializable()
class GetAllRequestsMessage(ImmediateSyftMessageWithReply):
    def __init__(
        self, address: Address, reply_to: Address, msg_id: Optional[UID] = None
    ):
        super().__init__(address=address, msg_id=msg_id, reply_to=reply_to)

    def _object2proto(self) -> GetAllRequestsMessage_PB:
        """Returns a protobuf serialization of self.

        As a requirement of all objects which inherit from Serializable,
        this method transforms the current object into the corresponding
        Protobuf object so that it can be further serialized.

        :return: returns a protobuf object
        :rtype: GetAllRequestsMessage_PB

        .. note::
            This method is purely an internal method. Please use serialize(object) or one of
            the other public serialization methods if you wish to serialize an
            object.
        """
        return GetAllRequestsMessage_PB(
            msg_id=serialize(self.id),
            address=serialize(self.address),
            reply_to=serialize(self.reply_to),
        )

    @staticmethod
    def _proto2object(proto: GetAllRequestsMessage_PB) -> "GetAllRequestsMessage":
        """Creates a ReprMessage from a protobuf

        As a requirement of all objects which inherit from Serializable,
        this method transforms a protobuf object into an instance of this class.

        :return: returns an instance of ReprMessage
        :rtype: ReprMessage

        .. note::
            This method is purely an internal method. Please use syft.deserialize()
            if you wish to deserialize an object.
        """

        return GetAllRequestsMessage(
            msg_id=_deserialize(blob=proto.msg_id),
            address=_deserialize(blob=proto.address),
            reply_to=_deserialize(blob=proto.reply_to),
        )

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        """Return the type of protobuf object which stores a class of this type

        As a part of serialization and deserialization, we need the ability to
        lookup the protobuf object type directly from the object type. This
        static method allows us to do this.

        Importantly, this method is also used to create the reverse lookup ability within
        the metaclass of Serializable. In the metaclass, it calls this method and then
        it takes whatever type is returned from this method and adds an attribute to it
        with the type of this class attached to it. See the MetaSerializable class for details.

        :return: the type of protobuf object which corresponds to this class.
        :rtype: GeneratedProtocolMessageType

        """

        return GetAllRequestsMessage_PB


@serializable()
class GetAllRequestsResponseMessage(ImmediateSyftMessageWithoutReply):
    def __init__(
        self,
        requests: List[RequestMessage],
        address: Address,
        msg_id: Optional[UID] = None,
    ):
        super().__init__(address=address, msg_id=msg_id)
        self.requests = requests

    def _object2proto(self) -> GetAllRequestsResponseMessage_PB:
        """Returns a protobuf serialization of self.

        As a requirement of all objects which inherit from Serializable,
        this method transforms the current object into the corresponding
        Protobuf object so that it can be further serialized.

        :return: returns a protobuf object
        :rtype: ReprMessage_PB

        .. note::
            This method is purely an internal method. Please use serialize(object) or one of
            the other public serialization methods if you wish to serialize an
            object.
        """
        return GetAllRequestsResponseMessage_PB(
            msg_id=serialize(self.id),
            address=serialize(self.address),
            requests=list(map(lambda x: serialize(x), self.requests)),
        )

    @staticmethod
    def _proto2object(
        proto: GetAllRequestsResponseMessage_PB,
    ) -> "GetAllRequestsResponseMessage":
        """Creates a GetAllRequestsResponseMessage from a protobuf

        As a requirement of all objects which inherit from Serializable,
        this method transforms a protobuf object into an instance of this class.

        :return: returns an instance of GetAllRequestsResponseMessage
        :rtype: GetAllRequestsResponseMessage

        .. note::
            This method is purely an internal method. Please use syft.deserialize()
            if you wish to deserialize an object.
        """

        return GetAllRequestsResponseMessage(
            msg_id=_deserialize(blob=proto.msg_id),
            address=_deserialize(blob=proto.address),
            requests=[_deserialize(blob=x) for x in proto.requests],
        )

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        """Return the type of protobuf object which stores a class of this type

        As a part of serialization and deserialization, we need the ability to
        lookup the protobuf object type directly from the object type. This
        static method allows us to do this.

        Importantly, this method is also used to create the reverse lookup ability within
        the metaclass of Serializable. In the metaclass, it calls this method and then
        it takes whatever type is returned from this method and adds an attribute to it
        with the type of this class attached to it. See the MetaSerializable class for details.

        :return: the type of protobuf object which corresponds to this class.
        :rtype: GeneratedProtocolMessageType

        """

        return GetAllRequestsResponseMessage_PB



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: proto/lib/python/range.proto
"""Generated protocol buffer code."""
# third party
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


# syft absolute
from syft.proto.core.common import (
    common_object_pb2 as proto_dot_core_dot_common_dot_common__object__pb2,
)

DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x1cproto/lib/python/range.proto\x12\x0fsyft.lib.python\x1a%proto/core/common/common_object.proto"U\n\x05Range\x12\r\n\x05start\x18\x01 \x01(\x03\x12\x0c\n\x04stop\x18\x02 \x01(\x03\x12\x0c\n\x04step\x18\x03 \x01(\x03\x12!\n\x02id\x18\x04 \x01(\x0b\x32\x15.syft.core.common.UIDb\x06proto3'
)


_RANGE = DESCRIPTOR.message_types_by_name["Range"]
Range = _reflection.GeneratedProtocolMessageType(
    "Range",
    (_message.Message,),
    {
        "DESCRIPTOR": _RANGE,
        "__module__": "proto.lib.python.range_pb2"
        # @@protoc_insertion_point(class_scope:syft.lib.python.Range)
    },
)
_sym_db.RegisterMessage(Range)

if _descriptor._USE_C_DESCRIPTORS == False:

    DESCRIPTOR._options = None
    _RANGE._serialized_start = 88
    _RANGE._serialized_end = 173
# @@protoc_insertion_point(module_scope)



# relative
from .new_user_messages import CreateUserMessage  # noqa: F401
from .new_user_messages import DeleteUserMessage  # noqa: F401
from .new_user_messages import GetUserMessage  # noqa: F401
from .new_user_messages import GetUsersMessage  # noqa: F401
from .new_user_messages import UpdateUserMessage  # noqa: F401



# stdlib
from typing import Optional

# third party
from google.protobuf.reflection import GeneratedProtocolMessageType

# syft absolute
import syft as sy

# relative
from ......proto.core.node.common.service.heritage_update_service_pb2 import (
    HeritageUpdateMessage as HeritageUpdateMessage_PB,
)
from .....common.message import ImmediateSyftMessageWithoutReply
from .....common.serde.serializable import serializable
from .....common.uid import UID
from .....io.address import Address


@serializable()
class HeritageUpdateMessage(ImmediateSyftMessageWithoutReply):
    def __init__(
        self,
        new_ancestry_address: Address,
        address: Address,
        msg_id: Optional[UID] = None,
    ):
        super().__init__(address=address, msg_id=msg_id)
        self.new_ancestry_address = new_ancestry_address

    def _object2proto(self) -> HeritageUpdateMessage_PB:
        return HeritageUpdateMessage_PB(
            new_ancestry_address=sy.serialize(self.new_ancestry_address),
            address=sy.serialize(self.address),
            msg_id=sy.serialize(self.id),
        )

    @staticmethod
    def _proto2object(proto: HeritageUpdateMessage_PB) -> "HeritageUpdateMessage":
        return HeritageUpdateMessage(
            new_ancestry_address=sy.deserialize(blob=proto.new_ancestry_address),
            address=sy.deserialize(blob=proto.address),
            msg_id=sy.deserialize(blob=proto.msg_id),
        )

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        return HeritageUpdateMessage_PB



# future
from __future__ import annotations

# stdlib
from collections.abc import Sequence
from functools import reduce
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union

# third party
from google.protobuf.reflection import GeneratedProtocolMessageType
import numpy as np
import numpy.typing as npt

# relative
from ....core.adp.entity import DataSubjectGroup as DSG
from ....core.adp.entity import Entity
from ....proto.core.adp.phi_tensor_pb2 import (
    RowEntityPhiTensor as RowEntityPhiTensor_PB,
)
from ....util import concurrency_count
from ....util import parallel_execution
from ....util import split_rows
from ...adp.vm_private_scalar_manager import VirtualMachinePrivateScalarManager
from ...common.serde.deserialize import _deserialize as deserialize
from ...common.serde.serializable import serializable
from ...common.serde.serialize import _serialize as serialize
from ...common.serde.types import Deserializeable
from ..broadcastable import is_broadcastable
from ..passthrough import AcceptableSimpleType  # type: ignore
from ..passthrough import PassthroughTensor  # type: ignore
from ..passthrough import implements  # type: ignore
from ..passthrough import is_acceptable_simple_type  # type: ignore
from .adp_tensor import ADPTensor
from .initial_gamma import InitialGammaTensor
from .intermediate_gamma import IntermediateGammaTensor as IGT
from .single_entity_phi import SingleEntityPhiTensor


def row_serialize(*rows: Any) -> List[Deserializeable]:
    return [serialize(row, to_bytes=True) for row in rows]


def row_deserialize(*rows: Deserializeable) -> List[Any]:
    output = []
    for row in rows:
        output.append(deserialize(row, from_bytes=True))
    return output


@serializable()
class RowEntityPhiTensor(PassthroughTensor, ADPTensor):
    """This tensor is one of several tensors whose purpose is to carry metadata
    relevant to automatically tracking the privacy budgets of tensor operations. This
    tensor is called 'Phi' tensor because it assumes that each number in the tensor
    originates from a single entity (no numbers originate from multiple entities). This
    tensor is called 'RowEntity' because it additionally assumes that all entries in a row
    come from the same entity (note: multiple rows may also be from the same or different
    entities). The reason we have a tensor specifically for tracking row-organized entities
    is that data entity-linked by row is very common and specifically accommodating it offers
    significant performance benefits over other DP tracking tensors. Note that when
    we refer to the number of 'rows' we simply refer to the length of the first dimension. This
    tensor can have an arbitrary number of dimensions."""

    def __init__(self, rows: Sequence, check_shape: bool = True):
        """Initialize a RowEntityPhiTensor

        rows: the actual data organized as an iterable (can be any type of iterable)
        check_shape: whether or not we are already confident that the objects in iterable
            'rows' all have the same dimension (check if we're not sure).

        """
        # Container type heirachy: https://docs.python.org/3/library/collections.abc.html
        self.child: Sequence
        super().__init__(rows)

        self.serde_concurrency: int = 0
        # include this check because it's expensive to check and sometimes we can skip it when
        # we already know the rows are identically shaped.
        if check_shape:

            # shape of the first row we use for reference
            shape = rows[0].shape

            # check each row to make sure it's the same shape as the first
            for row in rows[1:]:
                if shape != row.shape:
                    raise Exception(
                        f"All rows in RowEntityPhiTensor must match: {shape} != {row.shape}"
                    )

        """Calculate the number of unique entities behind the REPT"""
        self.unique_entities: set[Entity] = set()
        self.n_entities = 0
        for entity in self.entities.flatten():
            if isinstance(entity, str):
                entity = Entity(name=entity)

            if isinstance(entity, Entity):
                if entity not in self.unique_entities:
                    self.unique_entities.add(entity)
                    self.n_entities += 1
                else:
                    continue
            elif isinstance(entity, DSG):
                for e in entity.entity_set:
                    if e not in self.unique_entities:
                        self.unique_entities.add(e)
                        self.n_entities += 1
                    else:
                        continue
            else:
                raise Exception(f"{type(entity)}")

    @property
    def scalar_manager(self) -> VirtualMachinePrivateScalarManager:
        return self.child[0].scalar_manager

    @property
    def min_vals(self) -> np.ndarray:
        return np.concatenate([x.min_vals for x in self.child]).reshape(self.shape)

    @property
    def max_vals(self) -> np.ndarray:
        return np.concatenate([x.max_vals for x in self.child]).reshape(self.shape)

    @property
    def value(self) -> np.ndarray:
        return np.concatenate([x.child for x in self.child]).reshape(self.shape)

    @property
    def entities(self) -> np.ndarray:
        # we must cast the result of x.shape prod to an int because sometimes
        # it can be a float depending on the original type which fails to
        # multiply the [x.entity] array
        return np.array(
            [[x.entity] * int(np.array(x.shape).prod()) for x in self.child]
        ).reshape(self.shape)

    @property
    def gamma(self) -> InitialGammaTensor:
        return self.create_gamma()

    @property
    def dtype(self) -> np.dtype:
        # REPT child is a python List which does not have a np.dtype so we will return
        # the np.dtype of the first child in the row

        # TODO: We should decide what dtype an empty list has, numpy is float64
        if len(self.child) == 0:
            # we need to default to something
            return np.int32

        return self.child[0].dtype

    def astype(self, np_type: np.dtype) -> RowEntityPhiTensor:
        # RowEntityPhiTensor has a python List for its child
        return self.__class__(rows=[x.astype(np_type) for x in self.child])

    @staticmethod
    def convert_to_gamma(input_list: List) -> IGT:
        """This converts a REPT's data into a GammaTensor without having to initialize it. Used in comparison ops"""
        values = []
        entities = []
        mins = []
        maxes = []

        if len(input_list) > 1:
            target_shape = [len(input_list)] + list(input_list[0].shape)
        else:
            print(type(input_list), input_list)
            target_shape = input_list[0].shape

        for tensor in input_list:
            if isinstance(tensor, SingleEntityPhiTensor):
                values.append(tensor.child)
                entity_array = np.array(tensor.entity, dtype=object).repeat(
                    len(tensor.child.flatten())
                )
                entities.append(entity_array.reshape(tensor.shape))
                mins.append(tensor.min_vals)
                maxes.append(tensor.max_vals)
            elif isinstance(tensor, IGT):
                values.append(tensor._values())
                entities.append(tensor._entities())
                mins.append(tensor._min_values())
                maxes.append(tensor._max_values())
            else:
                raise Exception(f"Unknown type in REPT: {type(tensor)}")
        return InitialGammaTensor(
            values=np.concatenate(values).reshape(target_shape),
            entities=np.concatenate(entities).reshape(target_shape),
            min_vals=np.concatenate(mins).reshape(target_shape),
            max_vals=np.concatenate(maxes).reshape(target_shape),
            scalar_manager=input_list[0].scalar_manager,
        )

    def create_gamma(
        self, scalar_manager: Optional[VirtualMachinePrivateScalarManager] = None
    ) -> InitialGammaTensor:

        if scalar_manager is None:
            scalar_manager = self.scalar_manager

        return InitialGammaTensor(
            values=self.value,  # 5 x 10 data
            min_vals=self.min_vals,  # 5 x 10 minimum values
            max_vals=self.max_vals,  # 5 x 10 maximum values
            entities=self.entities,  # list of 5 entities
            scalar_manager=scalar_manager,
        )

    @property
    def shape(self) -> Tuple[Any, ...]:
        return tuple([len(self.child)] + list(self.child[0].shape))

    def __eq__(self, other: Any) -> Union[RowEntityPhiTensor, IGT]:

        if is_acceptable_simple_type(other) or len(self.child) == len(other.child):
            new_list = list()
            gamma_output = False
            for i in range(len(self.child)):
                if is_acceptable_simple_type(other):
                    new_list.append(self.child[i] == other)
                else:
                    result = self.child[i] == other.child[i]
                    if isinstance(result, InitialGammaTensor):
                        gamma_output = True
                    new_list.append(result)
            if not gamma_output:
                return RowEntityPhiTensor(rows=new_list, check_shape=False)
            else:
                return RowEntityPhiTensor.convert_to_gamma(new_list)
        else:
            raise Exception(
                f"Tensor dims do not match for __eq__: {len(self.child)} != {len(other.child)}"
            )

    def __ne__(self, other: Any) -> Union[RowEntityPhiTensor, IGT]:

        if is_acceptable_simple_type(other) or len(self.child) == len(other.child):
            new_list = list()
            gamma_output = False
            for i in range(len(self.child)):
                if is_acceptable_simple_type(other):
                    new_list.append(self.child[i] != other)
                else:
                    result = self.child[i] != other.child[i]
                    if isinstance(result, InitialGammaTensor):
                        gamma_output = True
                    new_list.append(result)
            if not gamma_output:
                return RowEntityPhiTensor(rows=new_list, check_shape=False)
            else:
                return RowEntityPhiTensor.convert_to_gamma(new_list)
        else:
            raise Exception(
                f"Tensor dims do not match for __ne__: {len(self.child)} != {len(other.child)}"
            )

    def __add__(
        self, other: Union[RowEntityPhiTensor, AcceptableSimpleType]
    ) -> RowEntityPhiTensor:
        # TODO: Catch unacceptable types (str, dict, etc) to avoid errors for other.child below
        if is_acceptable_simple_type(other) or len(self.child) == len(other.child):
            new_list = list()
            for i in range(len(self.child)):
                if is_acceptable_simple_type(other):
                    new_list.append(self.child[i] + other)
                else:
                    # Private/Public and Private/Private are handled by the underlying SEPT self.child objects.
                    new_list.append(self.child[i] + other.child[i])
            return RowEntityPhiTensor(rows=new_list, check_shape=False)
        else:
            # Broadcasting is possible, but we're skipping that for now.
            raise Exception(
                f"Tensor dims do not match for __add__: {len(self.child)} != {len(other.child)}"
            )

    def __sub__(
        self, other: Union[RowEntityPhiTensor, AcceptableSimpleType]
    ) -> RowEntityPhiTensor:
        # TODO: Catch unacceptable types (str, dict, etc) to avoid errors for other.child below
        if is_acceptable_simple_type(other) or len(self.child) == len(other.child):
            new_list = list()
            for i in range(len(self.child)):
                if is_acceptable_simple_type(other):
                    new_list.append(self.child[i] - other)
                else:
                    new_list.append(self.child[i] - other.child[i])
            return RowEntityPhiTensor(rows=new_list, check_shape=False)
        else:
            raise Exception(
                f"Tensor dims do not match for __sub__: {len(self.child)} != {len(other.child)}"
            )

    def __mul__(
        self, other: Union[RowEntityPhiTensor, AcceptableSimpleType]
    ) -> RowEntityPhiTensor:
        new_list = list()
        if is_acceptable_simple_type(other):
            if isinstance(other, np.ndarray):
                if is_broadcastable(self.shape, other.shape):
                    new_list.append(
                        [self.child[i] * other[i] for i in range(len(self.child))]
                    )
                else:
                    raise Exception(
                        f"Tensor dims do not match for __sub__: {getattr(self.child, 'shape', None)} != {other.shape}"
                    )
            else:  # int, float, bool, etc
                new_list = [child * other for child in self.child]
        elif isinstance(other, RowEntityPhiTensor):
            if is_broadcastable(self.shape, other.shape):
                for self_child, other_child in zip(self.child, other.child):
                    if len(self.child) != len(other.child):
                        raise ValueError(
                            "Zipping two different lengths will drop data. "
                            + f"{len(self.child)} vs {len(other.child)}"
                        )
                    new_list.append(self_child * other_child)
            else:
                raise Exception(
                    f"Tensor dims do not match for __sub__: {self.shape} != {other.shape}"
                )
        elif isinstance(other, SingleEntityPhiTensor):
            for child in self.child:
                # If even a single SEPT in the REPT isn't broadcastable, the multiplication operation doesn't work
                if not is_broadcastable(child.shape, other.shape):
                    raise Exception(
                        f"Tensor dims do not match for __sub__: {self.shape} != {other.shape}"
                    )
            new_list = [i * other for i in self.child]
        else:
            raise NotImplementedError
        return RowEntityPhiTensor(rows=new_list)

        # if is_acceptable_simple_type(other) or len(self.child) == len(other.child):
        #     new_list = list()
        #     for i in range(len(self.child)):
        #         if is_acceptable_simple_type(other):
        #             if isinstance(other, (int, bool, float)):
        #                 new_list.append(self.child[i] * other)
        #             else:
        #                 new_list.append(self.child[i] * other[i])
        #         else:
        #             if isinstance(other, RowEntityPhiTensor):
        #                 new_list.append(self.child[i] * other.child[i])
        #             elif isinstance(other, SingleEntityPhiTensor):
        #
        #                 new_list.append(self.child[i] * other)
        #     return RowEntityPhiTensor(rows=new_list, check_shape=False)
        # else:
        #     raise Exception(
        #         f"Tensor dims do not match for __mul__: {len(self.child)} != {len(other.child)}"
        #     )

    def __pos__(self) -> RowEntityPhiTensor:
        return RowEntityPhiTensor(rows=[+x for x in self.child], check_shape=False)

    def __neg__(self) -> RowEntityPhiTensor:
        return RowEntityPhiTensor(rows=[-x for x in self.child], check_shape=False)

    def __or__(self, other: Any) -> RowEntityPhiTensor:
        return RowEntityPhiTensor(
            rows=[x | other for x in self.child], check_shape=False
        )

    def __and__(self, other: Any) -> RowEntityPhiTensor:
        return RowEntityPhiTensor(
            rows=[x & other for x in self.child], check_shape=False
        )

    def __truediv__(
        self, other: Union[RowEntityPhiTensor, AcceptableSimpleType]
    ) -> RowEntityPhiTensor:

        if is_acceptable_simple_type(other) or len(self.child) == len(other.child):
            new_list = list()
            for i in range(len(self.child)):
                if is_acceptable_simple_type(other):
                    new_list.append(self.child[i] / other)
                else:
                    new_list.append(self.child[i] / other.child[i])
            return RowEntityPhiTensor(rows=new_list, check_shape=False)
        else:
            raise Exception(
                f"Tensor dims do not match for __truediv__: {len(self.child)} != {len(other.child)}"
            )

    def repeat(
        self, repeats: Union[int, List[int]], axis: Optional[int] = None
    ) -> RowEntityPhiTensor:

        if not isinstance(repeats, int):
            raise Exception(
                f"{type(self)}.repeat for repeats: List {repeats} not implemented yet"
            )

        if axis is None:
            raise Exception(
                "Conservatively, RowEntityPhiTensor doesn't yet support repeat(axis=None)"
            )

        if axis == 0 or axis == -len(self.shape):
            new_list = list()
            for _ in range(repeats):
                for row in self.child:
                    new_list.append(row)
            return RowEntityPhiTensor(rows=new_list, check_shape=False)

        elif axis > 0:
            new_list = list()
            for row in self.child:
                new_list.append(row.repeat(repeats, axis=axis - 1))
            return RowEntityPhiTensor(rows=new_list, check_shape=False)

        # axis is negative
        elif abs(axis) < len(self.shape):
            new_list = list()
            for row in self.child:
                new_list.append(row.repeat(repeats, axis=axis))
            return RowEntityPhiTensor(rows=new_list, check_shape=False)

        else:
            raise Exception(
                "'axis' arg is negative and strangely large... not sure what to do."
            )

    def flatten(self, order: Optional[str] = "C") -> RowEntityPhiTensor:
        new_list = list()
        for tensor in self.child:
            new_list.append(tensor.flatten(order))
        return RowEntityPhiTensor(rows=new_list, check_shape=False)

    def ravel(self, order: Optional[str] = "C") -> RowEntityPhiTensor:
        new_list = list()
        for tensor in self.child:
            new_list.append(tensor.ravel(order))
        return RowEntityPhiTensor(rows=new_list, check_shape=False)

    def swapaxes(self, axis1: int, axis2: int) -> RowEntityPhiTensor:
        if axis1 == 0 or axis2 == 0:
            raise Exception(
                "For now, you can't swap the first axis b/c that would "
                "probably create a Gamma Tensor. Sorry about that!"
            )
        new_list = list()
        for tensor in self.child:
            # Axis=0 for REPT.child is Axis=1 for REPT, so subtract 1
            new_list.append(tensor.swapaxes(axis1 - 1, axis2 - 1))
        return RowEntityPhiTensor(rows=new_list, check_shape=False)

    def squeeze(
        self, axis: Optional[Union[int, Tuple[int, ...]]] = None
    ) -> RowEntityPhiTensor:
        if axis == 0:
            # If the first axis can be squeezed then there is only one
            # tensor in the REPT, as such it might be a SEPT
            # TODO: Check if the output type is still a REPT
            # if isinstance(self.child[0], SEPT): return self.child[0]
            return RowEntityPhiTensor(rows=self.child[0])
        else:
            new_list = list()
            for tensor in self.child:
                new_list.append(tensor.squeeze(axis))
            self.child = new_list

        return RowEntityPhiTensor(rows=new_list, check_shape=False)

    def reshape(
        self,
        shape: Union[
            int,
            Union[Sequence[int], Sequence[Sequence[int]]],
        ],
    ) -> RowEntityPhiTensor:
        if isinstance(shape, int):
            raise Exception(
                f"{type(self)}.reshape for shape: int {shape} is not implemented"
            )
        # This is to fix the bug where shape = ([a, b, c], )
        if isinstance(shape[0], Sequence):
            shape = shape[0]

        if shape[0] != self.shape[0]:
            raise Exception(
                "For now, you can't reshape the first dimension because that would"
                + "probably require creating a gamma tensor."
                + str(shape)
                + " and "
                + str(self.shape)
            )

        new_list = list()
        for row in self.child:
            new_list.append(row.reshape(shape[1:]))

        return RowEntityPhiTensor(rows=new_list, check_shape=False)

    def resize(
        self,
        new_shape: Union[int, Tuple[int, ...]],
        refcheck: Optional[bool] = True,
    ) -> None:
        """This method is identical to reshape, but it modifies the Tensor in-place instead of returning a new one"""
        if isinstance(new_shape, int):
            raise Exception(f"new_shape: {new_shape} must be a Tuple for {type(self)}")
        if new_shape[0] != self.shape[0]:
            raise Exception(
                "For now, you can't reshape the first dimension because that would"
                + "probably require creating a gamma tensor."
            )

        new_list = list()
        for row in self.child:
            new_list.append(row.reshape(new_shape[1:]))

        # Modify the tensor data in-place instead of returning a new one.
        self.child = new_list

    def compress(
        self,
        condition: List[bool],
        axis: Optional[int] = None,
        out: Optional[np.ndarray] = None,
    ) -> RowEntityPhiTensor:
        # TODO: Could any conditions result in GammaTensors being formed?
        # TODO: Will min/max vals change upon filtering? I don't think so, since they're data independent
        new_list = list()
        for tensor in self.child:
            new_list.append(tensor.compress(condition, axis, out))
        return RowEntityPhiTensor(rows=new_list, check_shape=False)

    def partition(
        self,
        kth: Union[int, Tuple[int, ...]],
        axis: Optional[int] = -1,
        kind: Optional[str] = "introselect",
        order: Optional[Union[int, Tuple[int, ...]]] = None,
    ) -> RowEntityPhiTensor:
        if axis == 0:  # Unclear how to sort the SEPTs in a REPT
            raise NotImplementedError
        new_list = list()
        for tensor in self.child:
            new_list.append(tensor.partition(kth, axis, kind, order))
        return RowEntityPhiTensor(rows=new_list, check_shape=False)

    # Since this is being used differently compared to supertype, ignoring type annotation errors
    def pre_sum(self, *args: Any, **kwargs: Any) -> RowEntityPhiTensor:
        split_lst = []  # contains the different entities
        d = {}  # mapping of entities to list index
        c = 0  # to keep track of index count
        for i in self.child:
            if i.entity not in d:
                d[i.entity] = c
                split_lst.append([i])
                c += 1
            else:
                split_lst[d[i.entity]].append(i)

        # move to utils
        def list_sum(
            a: SingleEntityPhiTensor, b: SingleEntityPhiTensor
        ) -> SingleEntityPhiTensor:
            return a + b

        final_lst = []
        for i in split_lst:
            final_lst.append(reduce(list_sum, i))

        return RowEntityPhiTensor(final_lst)

    def sum(self, *args: Any, **kwargs: Any) -> IGT:
        # pre-sum sums all of the rows which are SEPT with the same
        # entity, this will reduce the number of input scalars for
        # the gamma tensor dramatically if there are any shared
        # entities. Final sum completes the sum, usually
        # returning a gammatensor.
        return self.pre_sum(*args, **kwargs).final_sum(*args, **kwargs)

    # Since this is being used differently compared to supertype, ignoring type annotation errors
    def final_sum(self, *args: Any, **kwargs: Any) -> IGT:
        # TODO: Check if this works if the number of dimensions/axes are passed as args/kwargs

        # pre-initialize result
        # target_shape = self.child[0].shape
        if len(args) > 0 or kwargs.get("axis", None) is not None:
            print(args)
            print(kwargs)
            raise NotImplementedError

        flat_symbols = []
        flat_values = []
        min_val_sum = 0
        max_val_sum = 0
        unique_entities = set()
        for row in self.child:
            if not isinstance(row, SingleEntityPhiTensor):
                raise NotImplementedError
            flat_child = row.child.flatten()
            flat_min = row.min_vals.flatten()
            flat_max = row.max_vals.flatten()

            flat_values.append(flat_child)

            min_val_sum += flat_min.sum()
            max_val_sum += flat_max.sum()
            for i in range(len(flat_child)):
                prime = self.scalar_manager.get_symbol(
                    min_val=flat_min[i],
                    value=flat_child[i],
                    max_val=flat_max[i],
                    entity=row.entity,
                )
                flat_symbols.append(prime)
            unique_entities.add(row.entity)

        term_tensor = (
            np.array(flat_symbols).reshape([1, len(flat_symbols)]).astype(np.int32)
        )
        coeff_tensor = np.ones_like(term_tensor)
        bias_tensor = np.zeros((1,), dtype=np.int32)
        value_tensor = np.sum(flat_values, axis=0)

        result = IGT(
            value_tensor=value_tensor,
            term_tensor=term_tensor,
            coeff_tensor=coeff_tensor,
            bias_tensor=bias_tensor,
            scalar_manager=self.scalar_manager,
            unique_entities=unique_entities,
        )
        result._min_vals_cache = min_val_sum
        result._max_vals_cache = max_val_sum

        return result

    # Since this is being used differently compared to supertype, ignoring type annotation errors
    def transpose(self, *dims: Optional[Any]) -> RowEntityPhiTensor:
        if dims:
            if dims[0] != 0:
                raise Exception("Can't move dim 0 in RowEntityPhiTensor at this time")

            new_dims = list(np.array(dims[1:]))

            new_list = list()
            for row in self.child:
                new_list.append(row.transpose(*new_dims))
        else:
            new_list = list()
            for row in self.child:
                new_list.append(row.transpose())
        return RowEntityPhiTensor(rows=new_list, check_shape=False)

    def __le__(self, other: Any) -> RowEntityPhiTensor:

        if is_acceptable_simple_type(other) or len(self.child) == len(other.child):
            new_list = list()
            gamma_output = False
            for i in range(len(self.child)):
                if is_acceptable_simple_type(other):
                    new_list.append(self.child[i] <= other)
                else:
                    result = self.child[i] <= other.child[i]
                    if isinstance(result, InitialGammaTensor):
                        gamma_output = True
                    new_list.append(result)
            if not gamma_output:
                return RowEntityPhiTensor(rows=new_list, check_shape=False)
            else:
                return RowEntityPhiTensor.convert_to_gamma(new_list)
        else:
            raise Exception(
                f"Tensor dims do not match for __le__: {len(self.child)} != {len(other.child)}"
            )

    def __lt__(self, other: Any) -> RowEntityPhiTensor:

        if is_acceptable_simple_type(other) or len(self.child) == len(other.child):
            new_list = list()
            gamma_output = False
            for i in range(len(self.child)):
                if is_acceptable_simple_type(other):
                    new_list.append(self.child[i] < other)
                else:
                    result = self.child[i] < other.child[i]
                    if isinstance(result, InitialGammaTensor):
                        gamma_output = True
                    new_list.append(result)
            if not gamma_output:
                return RowEntityPhiTensor(rows=new_list, check_shape=False)
            else:
                return RowEntityPhiTensor.convert_to_gamma(new_list)
        else:
            raise Exception(
                f"Tensor dims do not match for __lt__: {len(self.child)} != {len(other.child)}"
            )

    def __gt__(self, other: Any) -> RowEntityPhiTensor:

        if is_acceptable_simple_type(other) or len(self.child) == len(other.child):
            new_list = list()
            gamma_output = False
            for i in range(len(self.child)):
                if is_acceptable_simple_type(other):
                    new_list.append(self.child[i] > other)
                else:
                    result = self.child[i] > other.child[i]
                    if isinstance(result, InitialGammaTensor):
                        gamma_output = True
                    new_list.append(result)
            if not gamma_output:
                return RowEntityPhiTensor(rows=new_list, check_shape=False)
            else:
                return RowEntityPhiTensor.convert_to_gamma(new_list)
        else:
            raise Exception(
                f"Tensor dims do not match for __gt__: {len(self.child)} != {len(other.child)}"
            )

    def __ge__(self, other: Any) -> RowEntityPhiTensor:

        if is_acceptable_simple_type(other) or len(self.child) == len(other.child):
            new_list = list()
            gamma_output = False
            for i in range(len(self.child)):
                if is_acceptable_simple_type(other):
                    new_list.append(self.child[i] >= other)
                else:
                    result = self.child[i] >= other.child[i]
                    if isinstance(result, InitialGammaTensor):
                        gamma_output = True
                    new_list.append(result)
            if not gamma_output:
                return RowEntityPhiTensor(rows=new_list, check_shape=False)
            else:
                return RowEntityPhiTensor.convert_to_gamma(new_list)
        else:
            raise Exception(
                f"Tensor dims do not match for __ge__: {len(self.child)} != {len(other.child)}"
            )

    def cumprod(self, axis: Optional[int] = None) -> RowEntityPhiTensor:
        new_list = list()
        for tensor in self.child:
            new_list.append(tensor.cumprod(axis))
        return RowEntityPhiTensor(rows=new_list, check_shape=False)

    def cumsum(self, axis: Optional[int] = None) -> RowEntityPhiTensor:
        new_list = list()
        for tensor in self.child:
            new_list.append(tensor.cumsum(axis))
        return RowEntityPhiTensor(rows=new_list, check_shape=False)

    def clip(
        self, a_min: npt.ArrayLike, a_max: npt.ArrayLike, *args: Any
    ) -> RowEntityPhiTensor:

        if a_min is None and a_max is None:
            raise Exception("ValueError: clip: must set either max or min")

        new_list = list()
        for row in self.child:
            new_list.append(row.clip(a_min=a_min, a_max=a_max, *args))

        return RowEntityPhiTensor(rows=new_list, check_shape=False)

    def __floordiv__(
        self,
        other: Union[AcceptableSimpleType, SingleEntityPhiTensor, RowEntityPhiTensor],
    ) -> RowEntityPhiTensor:

        # We will let the underlying SingleEntityPhiTensor logic handle most of the errors/exceptions
        new_list = list()
        if is_acceptable_simple_type(other):
            for tensor in self.child:
                new_list.append(tensor // other)
        elif isinstance(other, SingleEntityPhiTensor):
            for tensor in self.child:
                new_list.append(tensor // other)
        elif isinstance(other, RowEntityPhiTensor):
            if not is_broadcastable(self.shape, other.shape):
                raise Exception(
                    f"Shapes not broadcastable: {self.shape}, {other.shape}"
                )
            else:
                if other.shape[0] == 1:
                    for tensor in self.child:
                        new_list.append(tensor // other.child[0])
                else:
                    if len(self.child) != len(other.child):
                        raise ValueError(
                            "Zipping two different lengths will drop data. "
                            + f"{len(self.child)} vs {len(other.child)}"
                        )
                    for self_tensors, other_tensors in zip(self.child, other.child):
                        new_list.append(self_tensors // other_tensors)
        else:
            raise NotImplementedError
        return RowEntityPhiTensor(rows=new_list, check_shape=False)

    def __mod__(
        self,
        other: Union[AcceptableSimpleType, SingleEntityPhiTensor, RowEntityPhiTensor],
    ) -> RowEntityPhiTensor:

        # We will let the underlying SingleEntityPhiTensor logic handle most of the errors/exceptions
        new_list = list()
        if is_acceptable_simple_type(other):
            for tensor in self.child:
                new_list.append(tensor % other)
        elif isinstance(other, SingleEntityPhiTensor):
            for tensor in self.child:
                new_list.append(tensor % other)
        elif isinstance(other, RowEntityPhiTensor):
            if not is_broadcastable(self.shape, other.shape):
                raise Exception(
                    f"Shapes not broadcastable: {self.shape}, {other.shape}"
                )
            else:
                if other.shape[0] == 1:
                    for tensor in self.child:
                        new_list.append(tensor % other.child[0])
                else:
                    if len(self.child) != len(other.child):
                        raise ValueError(
                            "Zipping two different lengths will drop data. "
                            + f"{len(self.child)} vs {len(other.child)}"
                        )
                    for self_tensors, other_tensors in zip(self.child, other.child):
                        new_list.append(self_tensors % other_tensors)
        else:
            raise NotImplementedError
        return RowEntityPhiTensor(rows=new_list, check_shape=False)

    def __divmod__(
        self,
        other: Union[AcceptableSimpleType, SingleEntityPhiTensor, RowEntityPhiTensor],
    ) -> Tuple:
        # Let logic written out in mod, floordiv and SEPT handle all exceptions
        return self // other, self % other

    def __matmul__(
        self,
        other: Union[AcceptableSimpleType, SingleEntityPhiTensor, RowEntityPhiTensor],
    ) -> RowEntityPhiTensor:
        new_list = list()
        if isinstance(other, np.ndarray):
            for tensor in self.child:
                new_list.append(tensor.__matmul__(other))
        elif isinstance(other, SingleEntityPhiTensor):
            # Whether the output of the matmul is SEPT or IGT, we let SEPT code determine that
            for tensor in self.child:
                new_list.append(tensor.__matmul__(other))
        elif isinstance(other, RowEntityPhiTensor):
            if len(self.child) != len(other.child):
                raise ValueError(
                    "Zipping two different lengths will drop data. "
                    + f"{len(self.child)} vs {len(other.child)}"
                )
            for self_tensor, other_tensor in zip(self.child, other.child):
                new_list.append(self_tensor.__matmul__(other_tensor))
        else:
            raise NotImplementedError
        return RowEntityPhiTensor(rows=new_list, check_shape=False)

    def trace(
        self,
        offset: int = 0,
        axis1: int = 1,
        axis2: int = 2,
        dtype: Optional[Any] = np.int32,
        out: Optional[np.ndarray] = None,
    ) -> RowEntityPhiTensor:
        if axis1 == 0 or axis2 == 0:
            raise NotImplementedError  # This would create a GammaTensor
        if dtype != np.int32:
            raise Exception(
                "We currently only support np.int32 dtypes for our tensors. "
                "We will be adding support for more dtypes soon! Sorry for the inconvenience."
            )

        # Axis #1 for REPT = Axis #0 for its SEPT child, etc
        axis1 -= 1
        axis2 -= 1
        new_list = list()
        for tensor in self.child:
            new_list.append(tensor.trace(offset, axis1, axis2, dtype, out))

        return RowEntityPhiTensor(rows=new_list, check_shape=False)

    def prod(
        self,
        axis: int = 1,  # might be bad that the default behaviour currently is not implemented...
        dtype: Optional[Any] = None,
        out: Optional[np.ndarray] = None,
        keepdims: Optional[bool] = False,
        initial: int = 1,
        where: Optional[bool] = True,
    ) -> RowEntityPhiTensor:
        if dtype and dtype != np.int32:
            raise Exception(
                "We currently only support np.int32 dtypes for our tensors. "
                "We will be adding support for more dtypes soon! Sorry for the inconvenience."
            )
        if axis == 0:
            raise NotImplementedError  # GammaTensor
        new_list = list()
        for tensor in self.child:
            new_list.append(tensor.prod(axis - 1, dtype, out, keepdims, initial, where))

        return RowEntityPhiTensor(rows=new_list, check_shape=False)

    def any(
        self,
        axis: Optional[int] = None,
        keepdims: Optional[bool] = False,
        where: Optional[bool] = True,
    ) -> RowEntityPhiTensor:
        """Test whether any element along a given axis evaluates to True"""

        new_list = list()
        for row in self.child:
            new_list.append(row.any(axis, keepdims, where))

        return RowEntityPhiTensor(rows=new_list, check_shape=False)

    def all(
        self,
        axis: Optional[int] = None,
        keepdims: Optional[bool] = False,
        where: Optional[bool] = True,
    ) -> RowEntityPhiTensor:
        """Test whether all elements along a given axis evaluates to True"""

        new_list = list()
        for row in self.child:
            new_list.append(row.all(axis, keepdims, where))

        return RowEntityPhiTensor(rows=new_list, check_shape=False)

    def abs(
        self,
        out: Optional[np.ndarray] = None,
    ) -> RowEntityPhiTensor:
        """Calculate the absolute value element-wise"""

        new_list = list()
        for row in self.child:
            new_list.append(row.abs(out))

        return RowEntityPhiTensor(rows=new_list, check_shape=False)

    def pow(
        self, value: Union[RowEntityPhiTensor, AcceptableSimpleType]
    ) -> RowEntityPhiTensor:
        """Return elements raised to powers from value, element-wise"""
        new_list = list()
        if is_acceptable_simple_type(value):
            if isinstance(value, np.ndarray):
                new_list.append(
                    [self.child[i].pow(value[i]) for i in range(len(self.child))]
                )
            else:  # int, float, bool, etc
                new_list = [child.pow(value) for child in self.child]
        elif isinstance(value, RowEntityPhiTensor):
            new_list = [
                self.child[i].pow(value.child[i]) for i in range(len(self.child))
            ]
        elif isinstance(value, SingleEntityPhiTensor):
            new_list = [i.pow(value) for i in self.child]
        else:
            raise NotImplementedError

        return RowEntityPhiTensor(rows=new_list)

    def copy(
        self, order: Optional[str] = "K", subok: Optional[bool] = True
    ) -> RowEntityPhiTensor:
        """Return copy of the given object"""
        new_list = list()
        for row in self.child:
            new_list.append(row.copy(order=order, subok=subok))

        return RowEntityPhiTensor(rows=new_list, check_shape=False)

    def take(
        self,
        indices: np.ArrayLike,
        axis: Optional[int] = None,
        mode: Optional[str] = "raise",
    ) -> RowEntityPhiTensor:
        """Take elements from an array along an axis"""
        new_list = list()
        for row in self.child:
            new_list.append(row.take(indices=indices, axis=axis, mode=mode))

        return RowEntityPhiTensor(rows=new_list, check_shape=False)

    def diagonal(
        self,
        offset: Optional[int] = 0,
        axis1: Optional[int] = 0,
        axis2: Optional[int] = 1,
    ) -> RowEntityPhiTensor:
        """Return specified diagonals"""
        new_list = list()
        for row in self.child:
            new_list.append(row.diagonal(offset=offset, axis1=axis1, axis2=axis2))

        return RowEntityPhiTensor(rows=new_list, check_shape=False)

    def round(self, decimals: int = 0) -> RowEntityPhiTensor:
        new_list = list()
        for tensor in self.child:
            new_list.append(tensor.round(decimals))

        return RowEntityPhiTensor(rows=new_list, check_shape=False)

    def _object2proto(self) -> RowEntityPhiTensor:
        entity_list = []
        entity_dict_index: Dict[Entity, int] = {}
        row_entity_index = []

        scalar_manager_list = []
        scalar_manager_dict_index: Dict[VirtualMachinePrivateScalarManager, int] = {}
        row_scalar_manager_index = []

        for i in self.child:
            entity = i.entity
            scalar_manager = i.scalar_manager

            if entity in entity_dict_index:
                index = entity_dict_index[entity]
            else:
                entity_list.append(entity)
                index = len(entity_list) - 1
                entity_dict_index[entity] = index
            row_entity_index.append(index)

            if scalar_manager in scalar_manager_dict_index:
                vm_index = scalar_manager_dict_index[scalar_manager]
            else:
                scalar_manager_list.append(scalar_manager)
                vm_index = len(scalar_manager_list) - 1
                scalar_manager_dict_index[scalar_manager] = vm_index
            row_scalar_manager_index.append(vm_index)
            i._remove_entity_scalar_manager = True

        if len(row_entity_index) != len(self.child):
            raise Exception("Length of entity index must match row length")

        if len(row_scalar_manager_index) != len(self.child):
            raise Exception("Length of scalar manager index must match row length")

        if self.serde_concurrency > 0 and concurrency_count() > 1:
            # serde_concurrency == 0 means off
            # serde_concurrency == 1 means auto detect cpu count
            # serde_concurrency >= 2 means manually set process count
            cpu_count = (
                self.serde_concurrency
                if self.serde_concurrency > 1
                else concurrency_count()
            )
            print(
                "Serializing with proto.serde_concurrency == ",
                self.serde_concurrency,
                "cpu_count",
                cpu_count,
            )
            args = split_rows(self.child, cpu_count=cpu_count)
            rows = parallel_execution(row_serialize, cpu_bound=bool(cpu_count))(args)
            output_rows = []
            for row in rows:
                output_rows.extend(row)
        else:
            output_rows = [serialize(row, to_bytes=True) for row in self.child]

        rept_pb = RowEntityPhiTensor_PB(
            serde_concurrency=int(self.serde_concurrency),
            rows=output_rows,
            unique_entities=[serialize(x) for x in entity_list],
            unique_scalar_managers=[serialize(x) for x in scalar_manager_list],
            row_entity_index=serialize(np.array(row_entity_index)),
            row_scalar_manager_index=serialize(np.array(row_scalar_manager_index)),
        )

        for child in self.child:
            del child._remove_entity_scalar_manager

        return rept_pb

    @staticmethod
    def _proto2object(proto: RowEntityPhiTensor_PB) -> RowEntityPhiTensor:
        # get back our entities and scalar managers
        unique_entities = [deserialize(x) for x in proto.unique_entities]
        unique_scalar_managers = [deserialize(x) for x in proto.unique_scalar_managers]

        row_entity_index = deserialize(proto.row_entity_index)
        row_scalar_manager_index = deserialize(proto.row_scalar_manager_index)

        if proto.serde_concurrency > 0 and concurrency_count() > 1:
            # serde_concurrency == 0 means off
            # serde_concurrency == 1 means auto detect cpu count
            # serde_concurrency >= 2 means manually set process count
            cpu_count = (
                proto.serde_concurrency
                if proto.serde_concurrency > 1
                else concurrency_count()
            )
            print(
                "Deserializing with proto.serde_concurrency == ",
                proto.serde_concurrency,
                "cpu_count",
                cpu_count,
            )
            args = split_rows(proto.rows, cpu_count)
            rows = parallel_execution(row_deserialize, cpu_bound=bool(cpu_count))(args)
            output_rows = []
            for row in rows:
                output_rows.extend(row)
        else:
            output_rows = [deserialize(row, from_bytes=True) for row in proto.rows]

        rows = []
        for i, row in enumerate(output_rows):
            row_index = row_entity_index[i]
            entity = unique_entities[row_index]
            scalar_manager_index = row_scalar_manager_index[i]
            scalar_manager = unique_scalar_managers[scalar_manager_index]

            # re-attach the original de-duplicated data before deserializing
            row.entity = entity
            row.scalar_manager = scalar_manager

            rows.append(row)

        rept = RowEntityPhiTensor(rows=rows)
        rept.serde_concurrency = proto.serde_concurrency
        return rept

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        return RowEntityPhiTensor_PB


@implements(RowEntityPhiTensor, np.expand_dims)
def expand_dims(a: np.typing.ArrayLike, axis: int) -> RowEntityPhiTensor:

    if axis == 0:
        raise Exception(
            "Currently, we don't have functionality for axis=0 but we could with a bit more work."
        )

    new_rows = list()
    for row in a.child:
        new_rows.append(np.expand_dims(row, axis - 1))

    return RowEntityPhiTensor(rows=new_rows, check_shape=False)



"""This module contains `EnumAttribute`, an AST node representing the
 attributes of a Python Enum which are only gettable, not settable."""

# stdlib
from enum import Enum
from typing import Any
from typing import Optional

# relative
from .. import ast
from ..core.common.pointer import AbstractPointer
from ..core.node.abstract.node import AbstractNodeClient
from ..core.node.common.action.get_enum_attribute_action import EnumAttributeAction
from ..logger import traceback_and_raise


class EnumAttribute(ast.attribute.Attribute):
    """An EnumAttribute represents the attributes of a Python Enum which are only gettable, not settable."""

    def __init__(
        self,
        parent: ast.attribute.Attribute,
        path_and_name: str,
        return_type_name: Optional[str] = None,
        client: Optional[AbstractNodeClient] = None,
    ) -> None:
        """Base constructor for Enum Attribute.

        Args:
            parent: The parent node is needed when solving `EnumAttributes`.
            path_and_name: The path for the current node, e.g. `syft.lib.python.List`.
            return_type_name: The return type name of given action as a string with its full path.
            client: The client for which all computation is being executed.
        """
        super().__init__(
            path_and_name=path_and_name,
            return_type_name=return_type_name,
            client=client,
            parent=parent,
        )

    def get_remote_enum_attribute(self) -> AbstractPointer:
        """Remote getter on an `Enum` attribute in the AST.

        Raises:
            ValueError: Cannot get Enum attribute.

        Returns:
            A pointer to the remote enum attribute.
        """
        if self.path_and_name is None:
            traceback_and_raise(
                ValueError(
                    "Can't get enum attribute, path_and_name to solve it "
                    "has not been set."
                )
            )

        if self.client is None:
            traceback_and_raise(
                ValueError(
                    "Can't get remote enum attribute if there is no client"
                    "set to get it from"
                )
            )

        return_tensor_type_pointer_type = self.client.lib_ast.query(
            path=self.return_type_name
        ).pointer_type

        ptr = return_tensor_type_pointer_type(client=self.client)

        msg = EnumAttributeAction(
            path=self.path_and_name,
            id_at_location=ptr.id_at_location,
            address=self.client.address,
        )
        self.client.send_immediate_msg_without_reply(msg=msg)
        return ptr

    def solve_get_enum_attribute(self) -> Enum:
        """Local getter on an `Enum` attribute in the AST.

        Raises:
            ValueError: Cannot get Enum attribute.

        Returns:
            The `Enum` object from the parent object reference.
        """
        if self.path_and_name is None:
            traceback_and_raise(
                ValueError(
                    "Can't get enum attribute, path_and_name to solve it "
                    "has not been set remotely."
                )
            )

        return getattr(self.parent.object_ref, self.name)

    def __call__(self, *args: Any, **kwargs: Any) -> None:
        """An `Enum` attribute is not callable.

        Args:
            *args: Variable length argument list.
            *kwargs: keyword arguments.

        Raises:
            ValueError: If the function is called.
        """
        traceback_and_raise(
            ValueError("__call__ should never be executed on an enum " "attribute.")
        )

    def add_path(self, *args: Any, **kwargs: Any) -> None:
        """An `Enum` can no longer have children nodes.

        Args:
            *args: Variable length argument list.
            *kwargs: keyword arguments..

        Raises:
            ValueError: If the function is called.
        """
        traceback_and_raise(
            "__add__path should never be called on an enum attribute, "
            "enum attributes are leaf nodes in the AST."
        )



# future
from __future__ import annotations

# stdlib
from typing import Optional
from typing import Union

# third party
from google.protobuf.reflection import GeneratedProtocolMessageType

# syft absolute
import syft as sy

# relative
from .....proto.core.adp.scalar_pb2 import BaseScalar as BaseScalar_PB
from ....common import UID
from ....common.serde.serializable import serializable
from ...entity import DataSubjectGroup
from ...entity import Entity
from .scalar import Scalar


@serializable()
class BaseScalar(Scalar):
    """
    A scalar which stores the root polynomial values. When this is a superclass of
    PhiScalar it represents data that was loaded in by a data owner. When this is a
    superclass of GammaScalar this represents the NODE at which point data from multiple
    entities was combined.
    """

    def __init__(
        self,
        min_val: float,
        value: Optional[float],
        max_val: float,
        entity: Optional[Union[Entity, DataSubjectGroup]] = None,
        id: Optional[UID] = None,
    ) -> None:
        self.id = id if id else UID()
        self._min_val: Optional[float] = float(min_val)
        self._value = float(value) if value is not None else None
        self._max_val: Optional[float] = float(max_val)
        self.entity = entity if entity is not None else Entity()

    @property
    def value(self) -> Optional[float]:
        return self._value

    @property
    def max_val(self) -> Optional[float]:
        return self._max_val

    @property
    def min_val(self) -> Optional[float]:
        return self._min_val

    def _object2proto(self) -> BaseScalar_PB:
        kwargs = {
            "id": sy.serialize(self.id, to_proto=True),
            "entity": sy.serialize(self.entity, to_proto=True),
        }

        for field in ["max_val", "min_val", "value"]:
            if getattr(self, field) is not None:
                kwargs[field] = getattr(self, field)
        pb = BaseScalar_PB(**kwargs)
        return pb

    @staticmethod
    def _proto2object(proto: BaseScalar_PB) -> BaseScalar:
        return BaseScalar(
            min_val=proto.min_val if proto.HasField("min_val") else None,
            max_val=proto.max_val if proto.HasField("max_val") else None,
            value=proto.value if proto.HasField("value") else None,
            entity=sy.deserialize(proto.entity, from_proto=True),
            id=sy.deserialize(proto.id, from_proto=True),
        )

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        return BaseScalar_PB



# stdlib
import time
from typing import Any
from typing import Dict as DictType
from typing import List
from typing import Optional

# third party
from nacl.signing import VerifyKey

# relative
from ......logger import debug
from ......logger import error
from ......util import traceback_and_raise
from ....abstract.node import AbstractNode
from ..node_service import ImmediateNodeServiceWithoutReply
from .request_handler_messages import GetAllRequestHandlersMessage
from .request_handler_messages import GetAllRequestHandlersResponseMessage
from .request_handler_messages import UpdateRequestHandlerMessage


class UpdateRequestHandlerService(ImmediateNodeServiceWithoutReply):
    @staticmethod
    def message_handler_types() -> List[type]:
        return [UpdateRequestHandlerMessage]

    @staticmethod
    def process(
        node: AbstractNode,
        msg: UpdateRequestHandlerMessage,
        verify_key: Optional[VerifyKey] = None,
    ) -> None:
        if verify_key is None:
            traceback_and_raise(
                ValueError(
                    "Can't process Request service without a given " "verification key"
                )
            )
        if verify_key == node.root_verify_key:
            replacement_handlers = []

            # find if there exists a handler match the handler passed in
            existing_handlers = getattr(node, "request_handlers", None)
            debug(f"> Updating Request Handlers with existing: {existing_handlers}")
            if existing_handlers is not None:
                matched = None
                for existing_handler in existing_handlers:
                    # we match two handlers according to their tags
                    if existing_handler["tags"] == msg.handler["tags"]:
                        matched = existing_handler
                        # if an existing_handler match the passed in handler,
                        # we ignore it in for loop
                        continue
                    else:
                        # if an existing_handler does not match the passed in
                        # handler, we keep it
                        replacement_handlers.append(existing_handler)

                if msg.keep:
                    msg.handler["created_time"] = time.time()
                    replacement_handlers.append(msg.handler)
                    if matched is not None:
                        debug(
                            f"> Replacing a Request Handler {matched} with: {msg.handler}"
                        )
                    else:
                        debug(f"> Adding a Request Handler {msg.handler}")
                else:
                    debug(f"> Removing a Request Handler with: {msg.handler}")

                setattr(node, "request_handlers", replacement_handlers)
                debug(f"> Finished Updating Request Handlers with: {existing_handlers}")
            else:
                error(f"> Node has no Request Handlers attribute: {type(node)}")

        return


class GetAllRequestHandlersService(ImmediateNodeServiceWithoutReply):
    @staticmethod
    def message_handler_types() -> List[type]:
        return [GetAllRequestHandlersMessage]

    @staticmethod
    def process(
        node: AbstractNode,
        msg: GetAllRequestHandlersMessage,
        verify_key: Optional[VerifyKey] = None,
    ) -> GetAllRequestHandlersResponseMessage:

        if verify_key is None:
            traceback_and_raise(
                ValueError(
                    "Can't process Request service without a given " "verification key"
                )
            )

        handlers: List[DictType[str, Any]] = []
        if verify_key == node.root_verify_key:
            existing_handlers = getattr(node, "request_handlers", None)
            debug(f"> Getting all Existing Request Handlers: {existing_handlers}")
            if existing_handlers is not None:
                handlers = existing_handlers

        return GetAllRequestHandlersResponseMessage(
            handlers=handlers, address=msg.reply_to
        )



# TODO: Unused Code
# Might use it later for VM case
"""
actions_lock = threading.Lock()
NodeSMPCAction = namedtuple("NodeSMPCAction", ["node_lock", "smpc_actions"])
actions_to_run_per_node: Dict[Any, NodeSMPCAction] = defaultdict(
    lambda: NodeSMPCAction(threading.Lock(), deque())
)


def consume_smpc_actions_round_robin() -> None:
    # Queue keeps a list of actions

    max_nr_retries = 10
    last_msg_id: Optional[UID] = None
    while 1:
        # Get a list of nodes
        with actions_lock:
            nodes = list(actions_to_run_per_node.keys())

        # Get one actions from each node in a Round Robin fashion and try to run it
        for node in nodes:
            with actions_to_run_per_node[node].node_lock:
                if len(actions_to_run_per_node[node].smpc_actions) == 0:
                    continue

                action = actions_to_run_per_node[node].smpc_actions[0]
                node, msg, verify_key, nr_retries = action
                if nr_retries > max_nr_retries:
                    raise ValueError(f"Retries to many times for {action}")

                try:
                    # try to execute and pop if succeded
                    msg.execute_action(node, verify_key)
                    actions_to_run_per_node[node].smpc_actions.popleft()
                except KeyError:
                    logger.warning(
                        f"Skip SMPC action {msg} since there was a key error when (probably) accessing the store"
                    )
                    if (last_msg_id is not None) and last_msg_id == msg.id:
                        # If there is only one action in all the lists
                        time.sleep(0.5)

                last_msg_id = msg.id


thread_smpc_action = threading.Thread(
    target=consume_smpc_actions_round_robin, args=(), daemon=True
)
thread_smpc_action.start()
"""



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: proto/grid/messages/request_messages.proto
"""Generated protocol buffer code."""
# third party
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


# syft absolute
from syft.proto.core.common import (
    common_object_pb2 as proto_dot_core_dot_common_dot_common__object__pb2,
)
from syft.proto.core.io import address_pb2 as proto_dot_core_dot_io_dot_address__pb2
from syft.proto.core.node.domain.service import (
    request_message_pb2 as proto_dot_core_dot_node_dot_domain_dot_service_dot_request__message__pb2,
)
from syft.proto.lib.python import dict_pb2 as proto_dot_lib_dot_python_dot_dict__pb2

DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n*proto/grid/messages/request_messages.proto\x12\x12syft.grid.messages\x1a%proto/core/common/common_object.proto\x1a\x1bproto/core/io/address.proto\x1a\x1bproto/lib/python/dict.proto\x1a\x34proto/core/node/domain/service/request_message.proto"\x9f\x01\n\x14\x43reateRequestMessage\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12&\n\x07\x61\x64\x64ress\x18\x02 \x01(\x0b\x32\x15.syft.core.io.Address\x12\x0f\n\x07\x63ontent\x18\x03 \x01(\t\x12\'\n\x08reply_to\x18\x04 \x01(\x0b\x32\x15.syft.core.io.Address"\x8c\x01\n\x15\x43reateRequestResponse\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12\x13\n\x0bstatus_code\x18\x02 \x01(\x05\x12\x0f\n\x07\x63ontent\x18\x03 \x01(\t\x12&\n\x07\x61\x64\x64ress\x18\x04 \x01(\x0b\x32\x15.syft.core.io.Address"\x8b\x01\n\x1a\x43reateBudgetRequestMessage\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12&\n\x07\x61\x64\x64ress\x18\x02 \x01(\x0b\x32\x15.syft.core.io.Address\x12\x0e\n\x06\x62udget\x18\x03 \x01(\x02\x12\x0e\n\x06reason\x18\x04 \x01(\t"\xa6\x01\n\x18GetBudgetRequestsMessage\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12&\n\x07\x61\x64\x64ress\x18\x02 \x01(\x0b\x32\x15.syft.core.io.Address\x12\x12\n\nrequest_id\x18\x03 \x01(\t\x12\'\n\x08reply_to\x18\x04 \x01(\x0b\x32\x15.syft.core.io.Address"\x92\x01\n\x19GetBudgetRequestsResponse\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12&\n\x07\x63ontent\x18\x03 \x03(\x0b\x32\x15.syft.lib.python.Dict\x12&\n\x07\x61\x64\x64ress\x18\x04 \x01(\x0b\x32\x15.syft.core.io.Address"\x9f\x01\n\x11GetRequestMessage\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12&\n\x07\x61\x64\x64ress\x18\x02 \x01(\x0b\x32\x15.syft.core.io.Address\x12\x12\n\nrequest_id\x18\x03 \x01(\t\x12\'\n\x08reply_to\x18\x04 \x01(\x0b\x32\x15.syft.core.io.Address"\x8c\x01\n\x12GetRequestResponse\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12\x13\n\x0bstatus_code\x18\x02 \x01(\x05\x12\x12\n\nrequest_id\x18\x03 \x01(\t\x12&\n\x07\x61\x64\x64ress\x18\x04 \x01(\x0b\x32\x15.syft.core.io.Address"\x8c\x01\n\x12GetRequestsMessage\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12&\n\x07\x61\x64\x64ress\x18\x02 \x01(\x0b\x32\x15.syft.core.io.Address\x12\'\n\x08reply_to\x18\x04 \x01(\x0b\x32\x15.syft.core.io.Address"\xa1\x01\n\x13GetRequestsResponse\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12\x13\n\x0bstatus_code\x18\x02 \x01(\x05\x12&\n\x07\x63ontent\x18\x03 \x03(\x0b\x32\x15.syft.lib.python.Dict\x12&\n\x07\x61\x64\x64ress\x18\x04 \x01(\x0b\x32\x15.syft.core.io.Address"\xa2\x01\n\x14\x44\x65leteRequestMessage\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12&\n\x07\x61\x64\x64ress\x18\x02 \x01(\x0b\x32\x15.syft.core.io.Address\x12\x12\n\nrequest_id\x18\x03 \x01(\t\x12\'\n\x08reply_to\x18\x04 \x01(\x0b\x32\x15.syft.core.io.Address"\x8f\x01\n\x15\x44\x65leteRequestResponse\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12\x13\n\x0bstatus_code\x18\x02 \x01(\x05\x12\x12\n\nrequest_id\x18\x03 \x01(\t\x12&\n\x07\x61\x64\x64ress\x18\x04 \x01(\x0b\x32\x15.syft.core.io.Address"\xb2\x01\n\x14UpdateRequestMessage\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12&\n\x07\x61\x64\x64ress\x18\x02 \x01(\x0b\x32\x15.syft.core.io.Address\x12\x0e\n\x06status\x18\x03 \x01(\t\x12\x12\n\nrequest_id\x18\x04 \x01(\t\x12\'\n\x08reply_to\x18\x05 \x01(\x0b\x32\x15.syft.core.io.Address"\x9f\x01\n\x15UpdateRequestResponse\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12\x13\n\x0bstatus_code\x18\x02 \x01(\x05\x12\x0e\n\x06status\x18\x03 \x01(\t\x12\x12\n\nrequest_id\x18\x04 \x01(\t\x12&\n\x07\x61\x64\x64ress\x18\x05 \x01(\x0b\x32\x15.syft.core.io.Address"\x8f\x01\n\x15GetAllRequestsMessage\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12&\n\x07\x61\x64\x64ress\x18\x02 \x01(\x0b\x32\x15.syft.core.io.Address\x12\'\n\x08reply_to\x18\x03 \x01(\x0b\x32\x15.syft.core.io.Address"\xaf\x01\n\x1dGetAllRequestsResponseMessage\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12&\n\x07\x61\x64\x64ress\x18\x02 \x01(\x0b\x32\x15.syft.core.io.Address\x12?\n\x08requests\x18\x03 \x03(\x0b\x32-.syft.core.node.domain.service.RequestMessageb\x06proto3'
)


_CREATEREQUESTMESSAGE = DESCRIPTOR.message_types_by_name["CreateRequestMessage"]
_CREATEREQUESTRESPONSE = DESCRIPTOR.message_types_by_name["CreateRequestResponse"]
_CREATEBUDGETREQUESTMESSAGE = DESCRIPTOR.message_types_by_name[
    "CreateBudgetRequestMessage"
]
_GETBUDGETREQUESTSMESSAGE = DESCRIPTOR.message_types_by_name["GetBudgetRequestsMessage"]
_GETBUDGETREQUESTSRESPONSE = DESCRIPTOR.message_types_by_name[
    "GetBudgetRequestsResponse"
]
_GETREQUESTMESSAGE = DESCRIPTOR.message_types_by_name["GetRequestMessage"]
_GETREQUESTRESPONSE = DESCRIPTOR.message_types_by_name["GetRequestResponse"]
_GETREQUESTSMESSAGE = DESCRIPTOR.message_types_by_name["GetRequestsMessage"]
_GETREQUESTSRESPONSE = DESCRIPTOR.message_types_by_name["GetRequestsResponse"]
_DELETEREQUESTMESSAGE = DESCRIPTOR.message_types_by_name["DeleteRequestMessage"]
_DELETEREQUESTRESPONSE = DESCRIPTOR.message_types_by_name["DeleteRequestResponse"]
_UPDATEREQUESTMESSAGE = DESCRIPTOR.message_types_by_name["UpdateRequestMessage"]
_UPDATEREQUESTRESPONSE = DESCRIPTOR.message_types_by_name["UpdateRequestResponse"]
_GETALLREQUESTSMESSAGE = DESCRIPTOR.message_types_by_name["GetAllRequestsMessage"]
_GETALLREQUESTSRESPONSEMESSAGE = DESCRIPTOR.message_types_by_name[
    "GetAllRequestsResponseMessage"
]
CreateRequestMessage = _reflection.GeneratedProtocolMessageType(
    "CreateRequestMessage",
    (_message.Message,),
    {
        "DESCRIPTOR": _CREATEREQUESTMESSAGE,
        "__module__": "proto.grid.messages.request_messages_pb2"
        # @@protoc_insertion_point(class_scope:syft.grid.messages.CreateRequestMessage)
    },
)
_sym_db.RegisterMessage(CreateRequestMessage)

CreateRequestResponse = _reflection.GeneratedProtocolMessageType(
    "CreateRequestResponse",
    (_message.Message,),
    {
        "DESCRIPTOR": _CREATEREQUESTRESPONSE,
        "__module__": "proto.grid.messages.request_messages_pb2"
        # @@protoc_insertion_point(class_scope:syft.grid.messages.CreateRequestResponse)
    },
)
_sym_db.RegisterMessage(CreateRequestResponse)

CreateBudgetRequestMessage = _reflection.GeneratedProtocolMessageType(
    "CreateBudgetRequestMessage",
    (_message.Message,),
    {
        "DESCRIPTOR": _CREATEBUDGETREQUESTMESSAGE,
        "__module__": "proto.grid.messages.request_messages_pb2"
        # @@protoc_insertion_point(class_scope:syft.grid.messages.CreateBudgetRequestMessage)
    },
)
_sym_db.RegisterMessage(CreateBudgetRequestMessage)

GetBudgetRequestsMessage = _reflection.GeneratedProtocolMessageType(
    "GetBudgetRequestsMessage",
    (_message.Message,),
    {
        "DESCRIPTOR": _GETBUDGETREQUESTSMESSAGE,
        "__module__": "proto.grid.messages.request_messages_pb2"
        # @@protoc_insertion_point(class_scope:syft.grid.messages.GetBudgetRequestsMessage)
    },
)
_sym_db.RegisterMessage(GetBudgetRequestsMessage)

GetBudgetRequestsResponse = _reflection.GeneratedProtocolMessageType(
    "GetBudgetRequestsResponse",
    (_message.Message,),
    {
        "DESCRIPTOR": _GETBUDGETREQUESTSRESPONSE,
        "__module__": "proto.grid.messages.request_messages_pb2"
        # @@protoc_insertion_point(class_scope:syft.grid.messages.GetBudgetRequestsResponse)
    },
)
_sym_db.RegisterMessage(GetBudgetRequestsResponse)

GetRequestMessage = _reflection.GeneratedProtocolMessageType(
    "GetRequestMessage",
    (_message.Message,),
    {
        "DESCRIPTOR": _GETREQUESTMESSAGE,
        "__module__": "proto.grid.messages.request_messages_pb2"
        # @@protoc_insertion_point(class_scope:syft.grid.messages.GetRequestMessage)
    },
)
_sym_db.RegisterMessage(GetRequestMessage)

GetRequestResponse = _reflection.GeneratedProtocolMessageType(
    "GetRequestResponse",
    (_message.Message,),
    {
        "DESCRIPTOR": _GETREQUESTRESPONSE,
        "__module__": "proto.grid.messages.request_messages_pb2"
        # @@protoc_insertion_point(class_scope:syft.grid.messages.GetRequestResponse)
    },
)
_sym_db.RegisterMessage(GetRequestResponse)

GetRequestsMessage = _reflection.GeneratedProtocolMessageType(
    "GetRequestsMessage",
    (_message.Message,),
    {
        "DESCRIPTOR": _GETREQUESTSMESSAGE,
        "__module__": "proto.grid.messages.request_messages_pb2"
        # @@protoc_insertion_point(class_scope:syft.grid.messages.GetRequestsMessage)
    },
)
_sym_db.RegisterMessage(GetRequestsMessage)

GetRequestsResponse = _reflection.GeneratedProtocolMessageType(
    "GetRequestsResponse",
    (_message.Message,),
    {
        "DESCRIPTOR": _GETREQUESTSRESPONSE,
        "__module__": "proto.grid.messages.request_messages_pb2"
        # @@protoc_insertion_point(class_scope:syft.grid.messages.GetRequestsResponse)
    },
)
_sym_db.RegisterMessage(GetRequestsResponse)

DeleteRequestMessage = _reflection.GeneratedProtocolMessageType(
    "DeleteRequestMessage",
    (_message.Message,),
    {
        "DESCRIPTOR": _DELETEREQUESTMESSAGE,
        "__module__": "proto.grid.messages.request_messages_pb2"
        # @@protoc_insertion_point(class_scope:syft.grid.messages.DeleteRequestMessage)
    },
)
_sym_db.RegisterMessage(DeleteRequestMessage)

DeleteRequestResponse = _reflection.GeneratedProtocolMessageType(
    "DeleteRequestResponse",
    (_message.Message,),
    {
        "DESCRIPTOR": _DELETEREQUESTRESPONSE,
        "__module__": "proto.grid.messages.request_messages_pb2"
        # @@protoc_insertion_point(class_scope:syft.grid.messages.DeleteRequestResponse)
    },
)
_sym_db.RegisterMessage(DeleteRequestResponse)

UpdateRequestMessage = _reflection.GeneratedProtocolMessageType(
    "UpdateRequestMessage",
    (_message.Message,),
    {
        "DESCRIPTOR": _UPDATEREQUESTMESSAGE,
        "__module__": "proto.grid.messages.request_messages_pb2"
        # @@protoc_insertion_point(class_scope:syft.grid.messages.UpdateRequestMessage)
    },
)
_sym_db.RegisterMessage(UpdateRequestMessage)

UpdateRequestResponse = _reflection.GeneratedProtocolMessageType(
    "UpdateRequestResponse",
    (_message.Message,),
    {
        "DESCRIPTOR": _UPDATEREQUESTRESPONSE,
        "__module__": "proto.grid.messages.request_messages_pb2"
        # @@protoc_insertion_point(class_scope:syft.grid.messages.UpdateRequestResponse)
    },
)
_sym_db.RegisterMessage(UpdateRequestResponse)

GetAllRequestsMessage = _reflection.GeneratedProtocolMessageType(
    "GetAllRequestsMessage",
    (_message.Message,),
    {
        "DESCRIPTOR": _GETALLREQUESTSMESSAGE,
        "__module__": "proto.grid.messages.request_messages_pb2"
        # @@protoc_insertion_point(class_scope:syft.grid.messages.GetAllRequestsMessage)
    },
)
_sym_db.RegisterMessage(GetAllRequestsMessage)

GetAllRequestsResponseMessage = _reflection.GeneratedProtocolMessageType(
    "GetAllRequestsResponseMessage",
    (_message.Message,),
    {
        "DESCRIPTOR": _GETALLREQUESTSRESPONSEMESSAGE,
        "__module__": "proto.grid.messages.request_messages_pb2"
        # @@protoc_insertion_point(class_scope:syft.grid.messages.GetAllRequestsResponseMessage)
    },
)
_sym_db.RegisterMessage(GetAllRequestsResponseMessage)

if _descriptor._USE_C_DESCRIPTORS == False:

    DESCRIPTOR._options = None
    _CREATEREQUESTMESSAGE._serialized_start = 218
    _CREATEREQUESTMESSAGE._serialized_end = 377
    _CREATEREQUESTRESPONSE._serialized_start = 380
    _CREATEREQUESTRESPONSE._serialized_end = 520
    _CREATEBUDGETREQUESTMESSAGE._serialized_start = 523
    _CREATEBUDGETREQUESTMESSAGE._serialized_end = 662
    _GETBUDGETREQUESTSMESSAGE._serialized_start = 665
    _GETBUDGETREQUESTSMESSAGE._serialized_end = 831
    _GETBUDGETREQUESTSRESPONSE._serialized_start = 834
    _GETBUDGETREQUESTSRESPONSE._serialized_end = 980
    _GETREQUESTMESSAGE._serialized_start = 983
    _GETREQUESTMESSAGE._serialized_end = 1142
    _GETREQUESTRESPONSE._serialized_start = 1145
    _GETREQUESTRESPONSE._serialized_end = 1285
    _GETREQUESTSMESSAGE._serialized_start = 1288
    _GETREQUESTSMESSAGE._serialized_end = 1428
    _GETREQUESTSRESPONSE._serialized_start = 1431
    _GETREQUESTSRESPONSE._serialized_end = 1592
    _DELETEREQUESTMESSAGE._serialized_start = 1595
    _DELETEREQUESTMESSAGE._serialized_end = 1757
    _DELETEREQUESTRESPONSE._serialized_start = 1760
    _DELETEREQUESTRESPONSE._serialized_end = 1903
    _UPDATEREQUESTMESSAGE._serialized_start = 1906
    _UPDATEREQUESTMESSAGE._serialized_end = 2084
    _UPDATEREQUESTRESPONSE._serialized_start = 2087
    _UPDATEREQUESTRESPONSE._serialized_end = 2246
    _GETALLREQUESTSMESSAGE._serialized_start = 2249
    _GETALLREQUESTSMESSAGE._serialized_end = 2392
    _GETALLREQUESTSRESPONSEMESSAGE._serialized_start = 2395
    _GETALLREQUESTSRESPONSEMESSAGE._serialized_end = 2570
# @@protoc_insertion_point(module_scope)



# third party
from sqlalchemy import Column
from sqlalchemy import DateTime
from sqlalchemy import Float
from sqlalchemy import ForeignKey
from sqlalchemy import Integer
from sqlalchemy import JSON
from sqlalchemy import String

# relative
from . import Base


class Request(Base):
    """Request.

    Columns:
        id (Integer, Primary Key): Cycle ID.
        date (TIME): Start time.
        user_id (Integer, Foreign Key): ID of the User that created the request.
        user_name (String): Email of the User that created the request.
        object_id (String): Target object to change in permisions.
        reason (String): Motivation of the request.
        status (String): The status of the request, wich can be 'pending', 'accepted' or 'denied'.
        request_type (String): Wheter the type of the request is 'permissions' or 'budget'.
        verify_key (String): User Verify Key.
    """

    __tablename__ = "request"

    id = Column(String(255), primary_key=True)
    date = Column(DateTime())
    user_id = Column(Integer, ForeignKey("syft_user.id"))
    user_name = Column(String(255))
    user_email = Column(String(255))
    user_role = Column(String(255))
    user_budget = Column(Float())
    institution = Column(String(255))
    website = Column(String(255))
    object_id = Column(String(255))
    reason = Column(String(255))
    status = Column(String(255), default="pending")
    request_type = Column(String(255))
    verify_key = Column(String(255))
    object_type = Column(String(255))
    tags = Column(JSON())
    updated_on = Column(DateTime())
    reviewer_name = Column(String(255))
    reviewer_role = Column(String(255))
    reviewer_comment = Column(String(255))
    requested_budget = Column(Float())
    current_budget = Column(Float())

    def __str__(self) -> str:
        return (
            f"< Request id : {self.id}, user: {self.user_id}, Date: {self.date}, Object: {self.object_id},"
            f" reason: {self.reason}, status: {self.status}, type: {self.object_type} >"
        )



# stdlib
import csv
import io
import tarfile
from typing import Callable
from typing import Dict as TypeDict
from typing import List
from typing import Type
from typing import Union

# third party
from nacl.signing import VerifyKey
import numpy as np
import torch as th

# relative
from ...... import deserialize
from ......util import get_tracer
from .....common.group import VERIFYALL
from .....common.message import ImmediateSyftMessageWithReply
from .....common.uid import UID
from .....store.storeable_object import StorableObject
from ....domain.domain_interface import DomainInterface
from ...exceptions import AuthorizationError
from ...exceptions import DatasetNotFoundError
from ...node_table.utils import model_to_json
from ..auth import service_auth
from ..node_service import ImmediateNodeServiceWithReply
from ..success_resp_message import SuccessResponseMessage
from .dataset_manager_messages import CreateDatasetMessage
from .dataset_manager_messages import DeleteDatasetMessage
from .dataset_manager_messages import GetDatasetMessage
from .dataset_manager_messages import GetDatasetResponse
from .dataset_manager_messages import GetDatasetsMessage
from .dataset_manager_messages import GetDatasetsResponse
from .dataset_manager_messages import UpdateDatasetMessage

ENCODING = "UTF-8"

tracer = get_tracer()


def _handle_dataset_creation_grid_ui(
    msg: CreateDatasetMessage, node: DomainInterface, verify_key: VerifyKey
) -> None:

    file_obj = io.BytesIO(msg.dataset)
    tar_obj = tarfile.open(fileobj=file_obj)
    tar_obj.extractall()
    dataset_id = node.datasets.register(**msg.metadata)
    for item in tar_obj.getmembers():
        if not item.isdir():
            extracted_file = tar_obj.extractfile(item.name)
            if not extracted_file:
                # TODO: raise CustomError
                raise ValueError("Dataset Tar corrupted")

            reader = csv.reader(
                extracted_file.read().decode().split("\n"),
                delimiter=",",
            )
            dataset = []

            for row in reader:
                if len(row) != 0:
                    dataset.append(row)
            dataset = np.array(dataset, dtype=np.float)
            df = th.tensor(dataset, dtype=th.float32)
            id_at_location = UID()

            # Step 2: create message which contains object to send
            storable = StorableObject(
                id=id_at_location,
                data=df,
                tags=["#" + item.name.split("/")[-1]],
                search_permissions={VERIFYALL: None},
                read_permissions={node.verify_key: node.id, verify_key: None},
                write_permissions={node.verify_key: node.id, verify_key: None},
            )
            node.store[storable.id] = storable

            node.datasets.add(
                name=item.name,
                dataset_id=str(dataset_id),
                obj_id=str(id_at_location.value),
                dtype=df.__class__.__name__,
                shape=str(tuple(df.shape)),
            )


def _handle_dataset_creation_syft(
    msg: CreateDatasetMessage, node: DomainInterface, verify_key: VerifyKey
) -> None:
    with tracer.start_as_current_span("_handle_dataset_creation_syft"):
        with tracer.start_as_current_span("deserialization"):
            result = deserialize(msg.dataset, from_bytes=True)
        dataset_id = msg.metadata.get("dataset_id")
        if not dataset_id:
            dataset_id = node.datasets.register(**msg.metadata)

        for table_name, table in result.items():
            id_at_location = UID()
            storable = StorableObject(
                id=id_at_location,
                data=table,
                tags=[f"#{table_name}"],
                search_permissions={VERIFYALL: None},
                read_permissions={node.verify_key: node.id, verify_key: None},
                write_permissions={node.verify_key: node.id, verify_key: None},
            )
            with tracer.start_as_current_span("save to DB"):
                node.store[storable.id] = storable

            node.datasets.add(
                name=table_name,
                dataset_id=str(dataset_id),
                obj_id=str(id_at_location.value),
                dtype=str(getattr(table, "dtype", type(table).__name__)),
                shape=str(table.shape),
            )


def create_dataset_msg(
    msg: CreateDatasetMessage,
    node: DomainInterface,
    verify_key: VerifyKey,
) -> SuccessResponseMessage:
    # Check key permissions
    _allowed = node.users.can_upload_data(verify_key=verify_key)

    if not _allowed:
        raise AuthorizationError("You're not allowed to upload data!")

    if msg.platform == "syft":
        _handle_dataset_creation_syft(msg, node, verify_key)
    elif msg.platform == "grid-ui":
        _handle_dataset_creation_grid_ui(msg, node, verify_key)

    return SuccessResponseMessage(
        address=msg.reply_to,
        resp_msg="Dataset Created Successfully!",
    )


def get_dataset_metadata_msg(
    msg: GetDatasetMessage,
    node: DomainInterface,
    verify_key: VerifyKey,
) -> GetDatasetResponse:
    ds, objs = node.datasets.get(msg.dataset_id)
    if not ds:
        raise DatasetNotFoundError
    dataset_json = model_to_json(ds)
    # these types seem broken
    dataset_json["data"] = [
        {"name": obj.name, "id": obj.obj, "dtype": obj.dtype, "shape": obj.shape}  # type: ignore
        for obj in objs
    ]
    return GetDatasetResponse(
        address=msg.reply_to,
        metadata=dataset_json,
    )


def get_all_datasets_metadata_msg(
    msg: GetDatasetsMessage,
    node: DomainInterface,
    verify_key: VerifyKey,
) -> GetDatasetsResponse:
    datasets = []
    for dataset in node.datasets.all():
        ds = model_to_json(dataset)
        _, objs = node.datasets.get(dataset.id)
        # these types seem broken
        ds["data"] = [
            {
                "name": obj.name,  # type: ignore
                "id": obj.obj,  # type: ignore
                "dtype": obj.dtype,  # type: ignore
                "shape": obj.shape,  # type: ignore
            }
            for obj in objs
        ]
        datasets.append(ds)
    return GetDatasetsResponse(
        address=msg.reply_to,
        metadatas=datasets,
    )


def update_dataset_msg(
    msg: UpdateDatasetMessage,
    node: DomainInterface,
    verify_key: VerifyKey,
) -> SuccessResponseMessage:
    # Get Payload Content
    _allowed = node.users.can_upload_data(verify_key=verify_key)
    if _allowed:
        metadata = {
            key: msg.metadata[key].upcast()
            for (key, value) in msg.metadata.items()
            if msg.metadata[key] is not None
        }

        node.datasets.set(dataset_id=msg.dataset_id, metadata=metadata)
    else:
        raise AuthorizationError("You're not allowed to upload data!")

    return SuccessResponseMessage(
        address=msg.reply_to,
        resp_msg="Dataset updated successfully!",
    )


def delete_dataset_msg(
    msg: DeleteDatasetMessage,
    node: DomainInterface,
    verify_key: VerifyKey,
) -> SuccessResponseMessage:
    _allowed = node.users.can_upload_data(verify_key=verify_key)

    if _allowed:
        # If bin object id exists, then only delete the bin object in the dataset.
        # Otherwise, delete the whole dataset.
        if msg.bin_object_id:
            key = UID(msg.bin_object_id)  # type: ignore
            node.store.delete(key)
        else:
            ds, objs = node.datasets.get(msg.dataset_id)

            if not ds:
                raise DatasetNotFoundError
            # Delete all the bin objects related to the dataset
            for obj in objs:
                node.store.delete(UID(obj.obj))  # type: ignore

            node.datasets.delete(id=msg.dataset_id)
    else:
        raise AuthorizationError("You're not allowed to delete data!")

    return SuccessResponseMessage(
        address=msg.reply_to,
        resp_msg="Dataset deleted successfully!",
    )


class DatasetManagerService(ImmediateNodeServiceWithReply):
    INPUT_TYPE = Union[
        Type[CreateDatasetMessage],
        Type[GetDatasetMessage],
        Type[GetDatasetsMessage],
        Type[UpdateDatasetMessage],
        Type[DeleteDatasetMessage],
    ]

    INPUT_MESSAGES = Union[
        CreateDatasetMessage,
        GetDatasetMessage,
        GetDatasetsMessage,
        UpdateDatasetMessage,
        DeleteDatasetMessage,
    ]

    OUTPUT_MESSAGES = Union[
        SuccessResponseMessage, GetDatasetResponse, GetDatasetsResponse
    ]

    msg_handler_map: TypeDict[INPUT_TYPE, Callable[..., OUTPUT_MESSAGES]] = {
        CreateDatasetMessage: create_dataset_msg,
        GetDatasetMessage: get_dataset_metadata_msg,
        GetDatasetsMessage: get_all_datasets_metadata_msg,
        UpdateDatasetMessage: update_dataset_msg,
        DeleteDatasetMessage: delete_dataset_msg,
    }

    @staticmethod
    @service_auth(guests_welcome=True)
    def process(
        node: DomainInterface,
        msg: INPUT_MESSAGES,
        verify_key: VerifyKey,
    ) -> OUTPUT_MESSAGES:
        return DatasetManagerService.msg_handler_map[type(msg)](
            msg=msg, node=node, verify_key=verify_key
        )

    @staticmethod
    def message_handler_types() -> List[Type[ImmediateSyftMessageWithReply]]:
        return [
            CreateDatasetMessage,
            GetDatasetMessage,
            GetDatasetsMessage,
            UpdateDatasetMessage,
            DeleteDatasetMessage,
        ]



# stdlib
from typing import Dict
from typing import List
from typing import Optional

# third party
from google.protobuf.reflection import GeneratedProtocolMessageType
from nacl.signing import VerifyKey

# syft absolute
import syft as sy

# relative
from .....proto.core.node.common.action.beaver_action_pb2 import (
    BeaverAction as BeaverAction_PB,
)
from ....common.serde.serializable import serializable
from ....common.uid import UID
from ....io.address import Address
from ....store.storeable_object import StorableObject
from ....tensor.smpc.share_tensor import ShareTensor
from ...abstract.node import AbstractNode
from .common import ImmediateActionWithoutReply

BEAVER_CACHE: Dict[UID, StorableObject] = {}  # Global cache for spdz mask values


@serializable()
class BeaverAction(ImmediateActionWithoutReply):
    def __init__(
        self,
        eps: ShareTensor,
        eps_id: UID,
        delta: ShareTensor,
        delta_id: UID,
        address: Address,
        msg_id: Optional[UID] = None,
    ):
        super().__init__(address=address, msg_id=msg_id)
        self.eps = eps
        self.eps_id = eps_id
        self.delta = delta
        self.delta_id = delta_id

    def __repr__(self) -> str:
        res = f"EPS: {self.eps}, "
        res = f"{res}EPS_ID: {self.eps_id}, "
        res = f"{res}DELTA: {self.delta}, "
        res = f"{res}DELTA_ID {self.delta_id} "
        return f"Beaver Action: {res}"

    @staticmethod
    def beaver_populate(
        data: ShareTensor, id_at_location: UID, node: AbstractNode
    ) -> None:
        """Populate the given input ShareTensor in the location specified.

        Args:
            data (Tensor): input ShareTensor to store in the node.
            id_at_location (UID): the location to store the data in.
            node Optional[AbstractNode] : The node on which the data is stored.
        """
        # TODO: Rasswanth Should modify storage to DB context,done
        # temporarily here to prevent race condition.

        obj = BEAVER_CACHE.get(id_at_location, None)  # type: ignore
        if obj is None:
            list_data = sy.lib.python.List([data])
            result = StorableObject(
                id=id_at_location,
                data=list_data,
                read_permissions={},
            )
            BEAVER_CACHE[id_at_location] = result  # type: ignore
        elif isinstance(obj.data, sy.lib.python.List):
            list_obj: List = obj.data
            list_obj.append(data)
            result = StorableObject(
                id=id_at_location,
                data=list_obj,
                read_permissions={},
            )
            BEAVER_CACHE[id_at_location] = result  # type: ignore
        else:
            raise Exception(f"Object at {id_at_location} should be a List or None")

    def execute_action(self, node: AbstractNode, verify_key: VerifyKey) -> None:
        eps = self.eps
        eps_id = self.eps_id
        delta = self.delta
        delta_id = self.delta_id
        BeaverAction.beaver_populate(eps, eps_id, node)
        BeaverAction.beaver_populate(delta, delta_id, node)

    def _object2proto(self) -> BeaverAction_PB:
        eps = sy.serialize(self.eps)
        eps_id = sy.serialize(self.eps_id)
        delta = sy.serialize(self.delta)
        delta_id = sy.serialize(self.delta_id)
        addr = sy.serialize(self.address)
        return BeaverAction_PB(
            eps=eps, eps_id=eps_id, delta=delta, delta_id=delta_id, address=addr
        )

    @staticmethod
    def _proto2object(proto: BeaverAction_PB) -> "BeaverAction":
        eps = sy.deserialize(blob=proto.eps)
        eps_id = sy.deserialize(blob=proto.eps_id)
        delta = sy.deserialize(blob=proto.delta)
        delta_id = sy.deserialize(blob=proto.delta_id)
        addr = sy.deserialize(blob=proto.address)
        return BeaverAction(
            eps=eps, eps_id=eps_id, delta=delta, delta_id=delta_id, address=addr
        )

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        return BeaverAction_PB



# future
from __future__ import annotations

# stdlib
from typing import Any
from typing import Optional
from typing import Tuple
from typing import Union

# third party
import numpy as np

# relative
from ..common.serde.serializable import serializable
from .entity import Entity


# allow us to serialize and deserialize np.arrays with strings inside as two np.arrays
# one containing the uint8 bytes and the other the offsets between strings
def numpyutf8tolist(string_index: Tuple[np.ndarray, np.ndarray]) -> np.ndarray:
    string_array, index_array = string_index
    output_bytes: bytes = string_array.astype(np.uint8).tobytes()
    output_list = []
    last_offset = 0
    for offset in index_array:
        chars = output_bytes[last_offset:offset]
        final_string = chars.decode("utf-8")
        last_offset = offset
        output_list.append(final_string)
    return np.array(output_list)


def liststrtonumpyutf8(string_list: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
    bytes_list = []
    indexes = []
    offset = 0
    for item in string_list:
        if not isinstance(item, (Entity, str)):
            raise Exception(
                f"EntityList entities must be List[Union[str, Entity]]. {type(item)}"
            )
        name = item if isinstance(item, str) else item.name
        name_bytes = name.encode("utf-8")
        offset += len(name_bytes)
        indexes.append(offset)
        bytes_list.append(name_bytes)

    np_bytes = np.frombuffer(b"".join(bytes_list), dtype=np.uint8)
    np_indexes = np.array(indexes)
    return (np_bytes, np_indexes)


@serializable(recursive_serde=True)
class EntityList:
    __attr_allowlist__ = ("one_hot_lookup", "entities_indexed")
    __slots__ = ("one_hot_lookup", "entities_indexed")

    # one_hot_lookup is a numpy array of unicode strings which can't be serialized
    __serde_overrides__ = {
        "one_hot_lookup": [liststrtonumpyutf8, numpyutf8tolist],
    }

    def __init__(
        self,
        one_hot_lookup: np.ndarray[Union[Entity, str]],
        entities_indexed: np.ndaray,
    ) -> None:
        self.one_hot_lookup = one_hot_lookup
        self.entities_indexed = entities_indexed

    @staticmethod
    def from_objs(entities: Union[np.ndarray, list]) -> EntityList:
        if isinstance(entities, list):
            entities = np.array(entities)
        one_hot_lookup, entities_indexed = np.unique(entities, return_inverse=True)

        return EntityList(one_hot_lookup, entities_indexed)

    def __getitem__(self, key: Union[int, slice, str]) -> Union[Entity, str]:
        return self.one_hot_lookup[self.entities_indexed[key]]

    def copy(self, order: Optional[str] = "K") -> EntityList:
        return EntityList(
            self.one_hot_lookup.copy(), self.entities_indexed.copy(order=order)
        )

    def __len__(self) -> int:
        return len(self.entities_indexed)

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, EntityList):
            if (self.entities_indexed == other.entities_indexed).all() and (  # type: ignore
                self.one_hot_lookup == other.one_hot_lookup
            ).all():
                return True
            return False
        return self == other



# stdlib
from typing import Any
from typing import List as TypeList
from typing import Optional
from typing import Union

# third party
import numpy as np

# relative
from ...adp.entity import DataSubjectGroup
from ...adp.vm_private_scalar_manager import VirtualMachinePrivateScalarManager
from ...common.serde.serializable import serializable
from ...common.uid import UID
from ..passthrough import PassthroughTensor  # type: ignore
from ..smpc.share_tensor import ShareTensor
from .adp_tensor import ADPTensor
from .intermediate_gamma import IntermediateGammaTensor


def numpy2list(np_obj: np.ndarray) -> TypeList:
    return [list(np_obj.flatten()), np_obj.shape]


def list2numpy(l_shape: Any) -> np.ndarray:
    list_length = l_shape[0]
    shape = l_shape[1]
    return np.array(list_length).reshape(shape)


@serializable(recursive_serde=True)
class InitialGammaTensor(IntermediateGammaTensor, ADPTensor):

    __attr_allowlist__ = [
        "uid",
        "values",
        "min_vals",
        "max_vals",
        "entities",
        "scalar_manager",
        "term_tensor",
        "coeff_tensor",
        "bias_tensor",
        "child",
    ]

    __serde_overrides__ = {"entities": [numpy2list, list2numpy]}

    sharetensor_values: Optional[ShareTensor]

    def __init__(
        self,
        values: Union[IntermediateGammaTensor, PassthroughTensor, np.ndarray],
        min_vals: np.ndarray,
        max_vals: np.ndarray,
        entities: Any,  # List[Entity] gives flatten errors
        scalar_manager: Optional[VirtualMachinePrivateScalarManager] = None,
    ) -> None:
        self.uid = UID()

        if isinstance(values, ShareTensor):
            self.sharetensor_values = values
            self.values = values.child
        else:
            self.sharetensor_values = None
            self.values = values

        self.min_vals = min_vals
        self._min_vals_cache = min_vals

        self.max_vals = max_vals
        self._max_vals_cache = max_vals

        self.entities = entities
        if scalar_manager is None:
            self.scalar_manager = VirtualMachinePrivateScalarManager()
        else:
            self.scalar_manager = scalar_manager

        flat_values = self.values.flatten()
        flat_min_vals = self.min_vals.flatten()
        flat_max_vals = self.max_vals.flatten()

        # If it's a list of lists, then it should still work
        if isinstance(self.entities, np.ndarray):
            flat_entities = self.entities.flatten()
        else:
            flat_entities = self.entities

        some_symbols = list()
        for i in range(flat_values.shape[0]):
            prime = self.scalar_manager.get_symbol(
                min_val=flat_min_vals[i],
                value=flat_values[i],
                max_val=flat_max_vals[i],
                entity=flat_entities[i]
                if not isinstance(flat_entities, DataSubjectGroup)
                else flat_entities,
            )
            some_symbols.append(prime)

        term_tensor = (
            np.array(some_symbols)
            .reshape(list(self.values.shape) + [1])
            .astype(np.int32)
        )
        coeff_tensor = (term_tensor * 0) + 1
        bias_tensor = self.values * 0

        if isinstance(entities, np.ndarray):
            unique_entities = set(list(entities.flatten()))
        else:
            unique_entities = set(entities)

        super().__init__(
            term_tensor=term_tensor,
            coeff_tensor=coeff_tensor,
            bias_tensor=bias_tensor,
            scalar_manager=self.scalar_manager,
            unique_entities=unique_entities,
        )
